from dataclasses import dataclass

from pydantic import Field, Secret

from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.processes.connectors.qdrant.qdrant import (
    QdrantAccessConfig,
    QdrantConnectionConfig,
    QdrantUploader,
    QdrantUploaderConfig,
    QdrantUploadStager,
    QdrantUploadStagerConfig,
)

CONNECTOR_TYPE = "qdrant-cloud"


class CloudQdrantAccessConfig(QdrantAccessConfig):
    api_key: str = Field(description="Qdrant API key")


class CloudQdrantConnectionConfig(QdrantConnectionConfig):
    url: str = Field(default=None, description="url of Qdrant Cloud")
    access_config: Secret[CloudQdrantAccessConfig]

    def get_client_kwargs(self) -> dict:
        return {
            "api_key": self.access_config.get_secret_value().api_key,
            "url": self.url,
        }


class CloudQdrantUploadStagerConfig(QdrantUploadStagerConfig):
    pass


@dataclass
class CloudQdrantUploadStager(QdrantUploadStager):
    upload_stager_config: CloudQdrantUploadStagerConfig


class CloudQdrantUploaderConfig(QdrantUploaderConfig):
    pass


@dataclass
class CloudQdrantUploader(QdrantUploader):
    connection_config: CloudQdrantConnectionConfig
    upload_config: CloudQdrantUploaderConfig
    connector_type: str = CONNECTOR_TYPE


qdrant_cloud_destination_entry = DestinationRegistryEntry(
    connection_config=CloudQdrantConnectionConfig,
    uploader=CloudQdrantUploader,
    uploader_config=CloudQdrantUploaderConfig,
    upload_stager=CloudQdrantUploadStager,
    upload_stager_config=CloudQdrantUploadStagerConfig,
)



import hashlib
import time
import xml.etree.ElementTree as ET
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import SourceConnectionError, ValueError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import SourceRegistryEntry
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from slack_sdk import WebClient
    from slack_sdk.web.async_client import AsyncWebClient

# NOTE: Pagination limit set to the upper end of the recommended range
# https://api.slack.com/apis/pagination#facts
PAGINATION_LIMIT = 200

CONNECTOR_TYPE = "slack"


class SlackAccessConfig(AccessConfig):
    token: str = Field(
        description="Bot token used to access Slack API, must have channels:history scope for the"
        " bot user."
    )


class SlackConnectionConfig(ConnectionConfig):
    access_config: Secret[SlackAccessConfig]

    @requires_dependencies(["slack_sdk"], extras="slack")
    @SourceConnectionError.wrap
    def get_client(self) -> "WebClient":
        from slack_sdk import WebClient

        return WebClient(token=self.access_config.get_secret_value().token)

    @requires_dependencies(["slack_sdk"], extras="slack")
    @SourceConnectionError.wrap
    def get_async_client(self) -> "AsyncWebClient":
        from slack_sdk.web.async_client import AsyncWebClient

        return AsyncWebClient(token=self.access_config.get_secret_value().token)


class SlackIndexerConfig(IndexerConfig):
    channels: list[str] = Field(
        description="Comma-delimited list of Slack channel IDs to pull messages from, can be"
        " both public or private channels."
    )
    start_date: Optional[datetime] = Field(
        default=None,
        description="Start date/time in formats YYYY-MM-DD[T]HH:MM[:SS[.ffffff]][Z or [±]HH[:]MM]"
        " or YYYY-MM-DD",
    )
    end_date: Optional[datetime] = Field(
        default=None,
        description="End date/time in formats YYYY-MM-DD[T]HH:MM[:SS[.ffffff]][Z or [±]HH[:]MM]"
        " or YYYY-MM-DD",
    )


@dataclass
class SlackIndexer(Indexer):
    index_config: SlackIndexerConfig
    connection_config: SlackConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        client = self.connection_config.get_client()
        for channel in self.index_config.channels:
            messages = []
            oldest = (
                str(self.index_config.start_date.timestamp())
                if self.index_config.start_date is not None
                else None
            )
            latest = (
                str(self.index_config.end_date.timestamp())
                if self.index_config.end_date is not None
                else None
            )
            for conversation_history in client.conversations_history(
                channel=channel,
                oldest=oldest,
                latest=latest,
                limit=PAGINATION_LIMIT,
            ):
                messages = conversation_history.get("messages", [])
                if messages:
                    yield self._messages_to_file_data(messages, channel)

    def _messages_to_file_data(
        self,
        messages: list[dict],
        channel: str,
    ) -> FileData:
        ts_oldest = min((message["ts"] for message in messages), key=lambda m: float(m))
        ts_newest = max((message["ts"] for message in messages), key=lambda m: float(m))

        identifier_base = f"{channel}-{ts_oldest}-{ts_newest}"
        identifier = hashlib.sha256(identifier_base.encode("utf-8")).hexdigest()
        filename = identifier[:16]

        source_identifiers = SourceIdentifiers(
            filename=f"{filename}.xml", fullpath=f"{filename}.xml"
        )
        return FileData(
            identifier=identifier,
            connector_type=CONNECTOR_TYPE,
            source_identifiers=source_identifiers,
            metadata=FileDataSourceMetadata(
                date_created=ts_oldest,
                date_modified=ts_newest,
                date_processed=str(time.time()),
                record_locator={
                    "channel": channel,
                    "oldest": ts_oldest,
                    "latest": ts_newest,
                },
            ),
            display_name=source_identifiers.fullpath,
        )

    @SourceConnectionError.wrap
    def precheck(self) -> None:
        client = self.connection_config.get_client()
        for channel in self.index_config.channels:
            # NOTE: Querying conversations history guarantees that the bot is in the channel
            client.conversations_history(channel=channel, limit=1)


class SlackDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class SlackDownloader(Downloader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: SlackConnectionConfig
    download_config: SlackDownloaderConfig = field(default_factory=SlackDownloaderConfig)

    def run(self, file_data, **kwargs):
        raise NotImplementedError

    async def run_async(self, file_data: FileData, **kwargs) -> DownloadResponse:
        # NOTE: Indexer should provide source identifiers required to generate the download path
        download_path = self.get_download_path(file_data)
        if download_path is None:
            logger.error(
                "Generated download path is None, source_identifiers might be missingfrom FileData."
            )
            raise ValueError("Generated invalid download path.")

        await self._download_conversation(file_data, download_path)
        return self.generate_download_response(file_data, download_path)

    def is_async(self):
        return True

    async def _download_conversation(self, file_data: FileData, download_path: Path) -> None:
        # NOTE: Indexer should supply the record locator in metadata
        if (
            file_data.metadata.record_locator is None
            or "channel" not in file_data.metadata.record_locator
            or "oldest" not in file_data.metadata.record_locator
            or "latest" not in file_data.metadata.record_locator
        ):
            logger.error(
                f"Invalid record locator in metadata: {file_data.metadata.record_locator}."
                "Keys 'channel', 'oldest' and 'latest' must be present."
            )
            raise ValueError("Invalid record locator.")

        client = self.connection_config.get_async_client()
        messages = []
        async for conversation_history in await client.conversations_history(
            channel=file_data.metadata.record_locator["channel"],
            oldest=file_data.metadata.record_locator["oldest"],
            latest=file_data.metadata.record_locator["latest"],
            limit=PAGINATION_LIMIT,
            # NOTE: In order to get the exact same range of messages as indexer, it provides
            # timestamps of oldest and newest messages, inclusive=True is necessary to include them
            inclusive=True,
        ):
            messages += conversation_history.get("messages", [])

        conversation = []
        for message in messages:
            thread_messages = []
            async for conversations_replies in await client.conversations_replies(
                channel=file_data.metadata.record_locator["channel"],
                ts=message["ts"],
                limit=PAGINATION_LIMIT,
            ):
                thread_messages += conversations_replies.get("messages", [])

            # NOTE: Replies contains the whole thread, including the message references by the `ts`
            # parameter even if it's the only message (there were no replies).
            # Reference: https://api.slack.com/methods/conversations.replies#markdown
            conversation.append(thread_messages)

        conversation_xml = self._conversation_to_xml(conversation)
        download_path.parent.mkdir(exist_ok=True, parents=True)
        conversation_xml.write(download_path, encoding="utf-8", xml_declaration=True)

    def _conversation_to_xml(self, conversation: list[list[dict]]) -> ET.ElementTree:
        root = ET.Element("messages")

        for thread in conversation:
            message, *replies = thread
            message_elem = ET.SubElement(root, "message")
            text_elem = ET.SubElement(message_elem, "text")
            text_elem.text = message.get("text")

            for reply in replies:
                reply_msg = reply.get("text", "")
                text_elem.text = "".join([str(text_elem.text), " <reply> ", reply_msg])

        return ET.ElementTree(root)


slack_source_entry = SourceRegistryEntry(
    indexer=SlackIndexer,
    indexer_config=SlackIndexerConfig,
    downloader=SlackDownloader,
    downloader_config=DownloaderConfig,
    connection_config=SlackConnectionConfig,
)



# https://developers.notion.com/reference/block#quote
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.attributes import Style
from htmlBuilder.tags import Div, HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import BlockBase
from unstructured_ingest.processes.connectors.notion.types.rich_text import RichText


@dataclass
class Quote(BlockBase):
    color: str
    children: List[dict] = field(default_factory=list)
    rich_text: List[RichText] = field(default_factory=list)

    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        rich_text = data.pop("rich_text", [])
        quote = cls(**data)
        quote.rich_text = [RichText.from_dict(rt) for rt in rich_text]
        return quote

    def get_html(self) -> Optional[HtmlTag]:
        if not self.rich_text:
            return None

        texts = [rt.get_html() for rt in self.rich_text]
        attributes = []
        if self.color and self.color != "default":
            attributes.append(Style(f"color: {self.color}"))
        return Div(attributes, texts)



from dataclasses import dataclass, field
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Generator, Optional
from urllib.parse import urlparse
from uuid import NAMESPACE_DNS, uuid5

from pydantic import Field, Secret, field_validator

from unstructured_ingest.data_types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import (
    ProviderError,
    UnstructuredIngestError,
    UserAuthError,
    UserError,
)
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    Indexer,
    IndexerConfig,
    download_responses,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    SourceRegistryEntry,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from github import ContentFile, GitTreeElement, Repository
    from github import Github as GithubClient
    from github.GithubException import GithubException
    from requests import HTTPError

CONNECTOR_TYPE = "github"


class GithubAccessConfig(AccessConfig):
    access_token: str = Field(description="Github acess token")


class GithubConnectionConfig(ConnectionConfig):
    access_config: Secret[GithubAccessConfig]
    url: str = Field(description="Github url or repository owner/name pair")

    @field_validator("url", mode="after")
    def conform_url(cls, value: str):
        parsed_url = urlparse(value)
        return parsed_url.path

    def get_full_url(self):
        return f"https://github.com/{self.url}"

    @requires_dependencies(["github"], extras="github")
    def get_client(self) -> "GithubClient":
        from github import Github as GithubClient

        return GithubClient(login_or_token=self.access_config.get_secret_value().access_token)

    def get_repo(self) -> "Repository":
        client = self.get_client()
        return client.get_repo(self.url)

    def wrap_github_exception(self, e: "GithubException") -> Exception:
        data = e.data
        status_code = e.status
        message = data.get("message")
        if status_code == 401:
            return UserAuthError(f"Unauthorized access to Github: {message}")
        if 400 <= status_code < 500:
            return UserError(message)
        if status_code > 500:
            return ProviderError(message)
        logger.debug(f"unhandled github error: {e}")
        return e

    def wrap_http_error(self, e: "HTTPError") -> Exception:
        status_code = e.response.status_code
        if status_code == 401:
            return UserAuthError(f"Unauthorized access to Github: {e.response.text}")
        if 400 <= status_code < 500:
            return UserError(e.response.text)
        if status_code > 500:
            return ProviderError(e.response.text)
        logger.debug(f"unhandled http error: {e}")
        return UnstructuredIngestError(str(e))

    @requires_dependencies(["requests"], extras="github")
    def wrap_error(self, e: Exception) -> Exception:
        from github.GithubException import GithubException
        from requests import HTTPError

        if isinstance(e, GithubException):
            return self.wrap_github_exception(e=e)
        if isinstance(e, HTTPError):
            return self.wrap_http_error(e=e)
        logger.debug(f"unhandled error: {e}")
        return UnstructuredIngestError(str(e))


class GithubIndexerConfig(IndexerConfig):
    branch: Optional[str] = Field(
        description="Branch to index, use the default if one isn't provided", default=None
    )
    recursive: bool = Field(
        description="Recursively index all files in the repository", default=True
    )


@dataclass
class GithubIndexer(Indexer):
    connection_config: GithubConnectionConfig
    index_config: GithubIndexerConfig = field(default_factory=GithubIndexerConfig)
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        try:
            self.connection_config.get_repo()
        except Exception as e:
            raise self.connection_config.wrap_error(e=e)

    def get_branch(self) -> str:
        repo = self.connection_config.get_repo()
        sha = self.index_config.branch or repo.default_branch
        return sha

    def list_files(self) -> list["GitTreeElement"]:
        repo = self.connection_config.get_repo()
        sha = self.index_config.branch or repo.default_branch
        git_tree = repo.get_git_tree(sha, recursive=self.index_config.recursive)
        file_elements = [
            element for element in git_tree.tree if element.size is not None and element.size > 0
        ]
        return file_elements

    def convert_element(self, element: "GitTreeElement") -> FileData:
        full_path = (
            f"{self.connection_config.get_full_url()}/blob/{self.get_branch()}/{element.path}"
        )

        return FileData(
            identifier=str(uuid5(NAMESPACE_DNS, full_path)),
            connector_type=self.connector_type,
            display_name=full_path,
            source_identifiers=SourceIdentifiers(
                filename=Path(element.path).name,
                fullpath=(Path(self.get_branch()) / element.path).as_posix(),
                rel_path=element.path,
            ),
            metadata=FileDataSourceMetadata(
                url=element.url,
                version=element.etag,
                record_locator={},
                date_modified=str(element.last_modified_datetime.timestamp()),
                date_processed=str(time()),
                filesize_bytes=element.size,
                permissions_data=[{"mode": element.mode}],
            ),
        )

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        for element in self.list_files():
            yield self.convert_element(element=element)


class GithubDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class GithubDownloader(Downloader):
    download_config: GithubDownloaderConfig
    connection_config: GithubConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["github"], extras="github")
    def get_file(self, file_data: FileData) -> "ContentFile":
        from github.GithubException import UnknownObjectException

        path = file_data.source_identifiers.relative_path
        repo = self.connection_config.get_repo()

        try:
            content_file = repo.get_contents(path)
        except UnknownObjectException as e:
            logger.error(f"File doesn't exists {self.connection_config.url}/{path}: {e}")
            raise UserError(f"File not found: {path}")
        return content_file

    @requires_dependencies(["requests"], extras="github")
    def get_contents(self, content_file: "ContentFile") -> bytes:
        import requests

        if content_file.decoded_content:
            return content_file.decoded_content
        download_url = content_file.download_url
        resp = requests.get(download_url)
        try:
            resp.raise_for_status()
        except requests.HTTPError as e:
            raise self.connection_config.wrap_error(e=e)
        return resp.content

    def run(self, file_data: FileData, **kwargs: Any) -> download_responses:
        content_file = self.get_file(file_data)
        contents = self.get_contents(content_file)
        download_path = self.get_download_path(file_data)
        with download_path.open("wb") as f:
            f.write(contents)
        return self.generate_download_response(file_data=file_data, download_path=download_path)


github_source_entry = SourceRegistryEntry(
    indexer=GithubIndexer,
    indexer_config=GithubIndexerConfig,
    downloader=GithubDownloader,
    downloader_config=GithubDownloaderConfig,
    connection_config=GithubConnectionConfig,
)



# https://developers.notion.com/reference/property-object#url
from dataclasses import dataclass, field
from typing import Optional

from htmlBuilder.attributes import Href
from htmlBuilder.tags import A, HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import DBCellBase, DBPropertyBase


@dataclass
class URL(DBPropertyBase):
    id: str
    name: str
    type: str = "url"
    url: dict = field(default_factory=dict)
    description: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class URLCell(DBCellBase):
    id: str
    url: Optional[str] = None
    name: Optional[str] = None
    type: str = "url"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        if url := self.url:
            return A([Href(url)], url)
        return None



# https://developers.notion.com/reference/property-object#formula
from dataclasses import dataclass
from typing import Optional

from htmlBuilder.tags import Div, HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import (
    DBCellBase,
    DBPropertyBase,
    FromJSONMixin,
)


@dataclass
class FormulaProp(FromJSONMixin):
    expression: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class Formula(DBPropertyBase):
    id: str
    name: str
    formula: FormulaProp
    type: str = "formula"
    description: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(formula=FormulaProp.from_dict(data.pop("formula", {})), **data)


@dataclass
class FormulaCell(DBCellBase):
    id: str
    formula: dict
    type: str = "formula"
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        formula = self.formula
        t = formula.get("type")
        return Div([], str(formula[t]))



import json
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Generator, Optional

from dateutil import parser
from pydantic import Field, Secret
from pydantic.functional_validators import BeforeValidator

from unstructured_ingest.data_types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import SourceConnectionError, UserAuthError, ValueError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import SourceRegistryEntry
from unstructured_ingest.processes.connectors.utils import conform_string_to_dict
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from googleapiclient.discovery import Resource as GoogleAPIResource

CONNECTOR_TYPE = "google_drive"


# Maps Google-native Drive MIME types → export MIME types
GOOGLE_EXPORT_MIME_MAP = {
    "application/vnd.google-apps.document": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",  # noqa: E501
    "application/vnd.google-apps.spreadsheet": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",  # noqa: E501
    "application/vnd.google-apps.presentation": "application/vnd.openxmlformats-officedocument.presentationml.presentation",  # noqa: E501
}

# Maps export MIME types → file extensions
EXPORT_EXTENSION_MAP = {
    "application/vnd.openxmlformats-officedocument.wordprocessingml.document": ".docx",
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": ".xlsx",
    "application/vnd.openxmlformats-officedocument.presentationml.presentation": ".pptx",
    "application/pdf": ".pdf",
    "text/html": ".html",
}

# LRO Export Size Threshold is 10MB in real but the exported file might be slightly larger
# than the original Google Workspace file - thus the threshold is set to 9MB
LRO_EXPORT_SIZE_THRESHOLD = 9 * 1024 * 1024  # 9MB


class GoogleDriveAccessConfig(AccessConfig):
    service_account_key: Optional[Annotated[dict, BeforeValidator(conform_string_to_dict)]] = Field(
        default=None, description="Credentials values to use for authentication"
    )
    service_account_key_path: Optional[Path] = Field(
        default=None,
        description="File path to credentials values to use for authentication",
    )

    def model_post_init(self, __context: Any) -> None:
        if self.service_account_key is None and self.service_account_key_path is None:
            raise ValueError(
                "either service_account_key or service_account_key_path must be provided"
            )

    def get_service_account_key(self) -> dict:
        key_data = None
        if self.service_account_key_path:
            with self.service_account_key_path.open() as f:
                key_data = json.load(f)
        if key_data and self.service_account_key:
            if key_data == self.service_account_key:
                return key_data
            else:
                raise ValueError(
                    "service_account_key and service_account_key_path "
                    "both provided and have different values"
                )
        if key_data:
            return key_data
        return self.service_account_key


class GoogleDriveConnectionConfig(ConnectionConfig):
    drive_id: str = Field(description="Google Drive File or Folder ID.")
    access_config: Secret[GoogleDriveAccessConfig]

    @requires_dependencies(["googleapiclient"], extras="google-drive")
    @contextmanager
    def get_client(self) -> Generator["GoogleAPIResource", None, None]:
        from google.auth import exceptions
        from google.oauth2 import service_account
        from googleapiclient.discovery import build
        from googleapiclient.errors import HttpError

        access_config = self.access_config.get_secret_value()
        key_data = access_config.get_service_account_key()

        try:
            creds = service_account.Credentials.from_service_account_info(key_data)
            service = build("drive", "v3", credentials=creds)
            with service.files() as client:
                yield client
        except HttpError as exc:
            raise ValueError(f"{exc.reason}")
        except exceptions.DefaultCredentialsError:
            raise UserAuthError("The provided API key is invalid.")


class GoogleDriveIndexerConfig(IndexerConfig):
    extensions: Optional[list[str]] = None
    recursive: bool = False

    def model_post_init(self, __context: Any) -> None:
        if self.extensions is not None:
            self.extensions = [e.lstrip(".") for e in self.extensions]


@dataclass
class GoogleDriveIndexer(Indexer):
    connection_config: GoogleDriveConnectionConfig
    index_config: GoogleDriveIndexerConfig
    fields: list[str] = field(
        default_factory=lambda: [
            "id",
            "name",
            "mimeType",
            "fileExtension",
            "md5Checksum",
            "sha1Checksum",
            "sha256Checksum",
            "headRevisionId",
            "permissions",
            "createdTime",
            "modifiedTime",
            "version",
            "originalFilename",
            "capabilities",
            "permissionIds",
            "size",
        ]
    )

    @staticmethod
    def verify_drive_api_enabled(client) -> None:
        from googleapiclient.errors import HttpError

        """
        Makes a lightweight API call to verify that the Drive API is enabled.
        If the API is not enabled, an HttpError should be raised.
        """
        try:
            # A very minimal call: list 1 file from the drive.
            client.list(
                supportsAllDrives=True,
                includeItemsFromAllDrives=True,
                spaces="drive",
                pageSize=1,
                fields="files(id)",
            ).execute()
        except HttpError as e:
            error_content = e.content.decode() if hasattr(e, "content") else ""
            lower_error = error_content.lower()
            if "drive api" in lower_error and (
                "not enabled" in lower_error or "not been used" in lower_error
            ):
                raise SourceConnectionError(
                    "Google Drive API is not enabled for your project. \
                    Please enable it in the Google Cloud Console."
                )
            else:
                raise SourceConnectionError("Google drive API unreachable for an unknown reason!")

    @staticmethod
    def count_files_recursively(
        files_client: "GoogleAPIResource", folder_id: str, extensions: list[str] = None
    ) -> int:
        """
        Count non-folder files recursively under the given folder.
        If `extensions` is provided, only count files
        whose `fileExtension` matches one of the values.
        """
        count = 0
        stack = [folder_id]
        # Pre-compute lower-case extension set for O(1) lookup
        valid_exts = set(e.lower() for e in extensions) if extensions else None

        while stack:
            current_folder = stack.pop()
            # Always list all items under the current folder.
            query = f"'{current_folder}' in parents"
            page_token = None
            while True:
                response = files_client.list(
                    supportsAllDrives=True,
                    includeItemsFromAllDrives=True,
                    spaces="drive",
                    q=query,
                    fields="nextPageToken, files(id, mimeType, fileExtension)",
                    pageToken=page_token,
                    pageSize=1000,
                ).execute()
                for item in response.get("files", []):
                    if item.get("mimeType") == "application/vnd.google-apps.folder":
                        # Always traverse sub-folders regardless of extension filter.
                        stack.append(item["id"])
                    else:
                        if extensions:
                            # Use a case-insensitive comparison for the file extension.
                            file_ext = (item.get("fileExtension") or "").lower()
                            if file_ext in valid_exts:
                                count += 1
                        else:
                            count += 1
                page_token = response.get("nextPageToken")
                if not page_token:
                    break
        return count

    def precheck(self) -> None:
        """
        Enhanced precheck that verifies not only connectivity
        but also that the provided drive_id is valid and accessible.
        """
        try:
            with self.connection_config.get_client() as client:
                # First, verify that the Drive API is enabled.
                self.verify_drive_api_enabled(client)

                # Try to retrieve metadata for the drive id.
                # This will catch errors such as an invalid drive id or insufficient permissions.
                root_info = self.get_root_info(
                    files_client=client, object_id=self.connection_config.drive_id
                )
                logger.info(
                    f"Successfully retrieved drive root info: "
                    f"{root_info.get('name', 'Unnamed')} (ID: {root_info.get('id')})"
                )

            # If the target is a folder, perform file count check.
            if self.is_dir(root_info):
                if self.index_config.recursive:
                    file_count = self.count_files_recursively(
                        client,
                        self.connection_config.drive_id,
                        extensions=self.index_config.extensions,
                    )
                    if file_count == 0:
                        logger.warning(
                            "Empty folder: no files found recursively in the folder. \
                             Please verify that the folder contains files and \
                             that the service account has proper permissions."
                        )
                        # raise SourceConnectionError(
                        #     "Empty folder: no files found recursively in the folder. "
                        #     "Please verify that the folder contains files and \
                        #     that the service account has proper permissions."
                        # )
                    else:
                        logger.info(f"Found {file_count} files recursively in the folder.")
                else:
                    # Non-recursive: check for at least one immediate non-folder child.
                    response = client.list(
                        supportsAllDrives=True,
                        includeItemsFromAllDrives=True,
                        spaces="drive",
                        fields="files(id)",
                        pageSize=1,
                        q=f"'{self.connection_config.drive_id}' in parents",
                    ).execute()
                    if not response.get("files"):
                        logger.warning(
                            "Empty folder: no files found at the folder's root level. "
                            "Please verify that the folder contains files and \
                            that the service account has proper permissions."
                        )
                        # raise SourceConnectionError(
                        #     "Empty folder: no files found at the folder's root level. "
                        #     "Please verify that the folder contains files and \
                        #     that the service account has proper permissions."
                        # )
                    else:
                        logger.info("Found files at the folder's root level.")
            else:
                # If the target is a file, precheck passes.
                logger.info("Drive ID corresponds to a file. Precheck passed.")

        except Exception as e:
            logger.error(
                "Failed to validate Google Drive connection during precheck",
                exc_info=True,
            )
            raise SourceConnectionError(f"Precheck failed: {e}")

    @staticmethod
    def is_dir(record: dict) -> bool:
        return record.get("mimeType") == "application/vnd.google-apps.folder"

    @staticmethod
    def map_file_data(root_info: dict) -> FileData:
        file_id = root_info["id"]
        filename = root_info.pop("name")
        url = root_info.pop("webContentLink", None)
        version = root_info.pop("version", None)
        permissions = root_info.pop("permissions", None)
        date_created_str = root_info.pop("createdTime", None)
        date_created_dt = parser.parse(date_created_str) if date_created_str else None
        date_modified_str = root_info.pop("modifiedTime", None)
        parent_path = root_info.pop("parent_path", None)
        parent_root_path = root_info.pop("parent_root_path", None)
        date_modified_dt = parser.parse(date_modified_str) if date_modified_str else None
        if (
            parent_path
            and isinstance(parent_path, str)
            and parent_root_path
            and isinstance(parent_root_path, str)
        ):
            fullpath = f"{parent_path}/{filename}"
            rel_path = Path(fullpath).relative_to(parent_root_path).as_posix()
            source_identifiers = SourceIdentifiers(
                filename=filename, fullpath=fullpath, rel_path=rel_path
            )
        else:
            source_identifiers = SourceIdentifiers(fullpath=filename, filename=filename)
        return FileData(
            connector_type=CONNECTOR_TYPE,
            identifier=file_id,
            source_identifiers=source_identifiers,
            metadata=FileDataSourceMetadata(
                url=url,
                version=version,
                date_created=str(date_created_dt.timestamp()),
                date_modified=str(date_modified_dt.timestamp()),
                permissions_data=permissions,
                record_locator={
                    "file_id": file_id,
                },
            ),
            additional_metadata=root_info,
            display_name=source_identifiers.fullpath,
        )

    def get_paginated_results(
        self,
        files_client,
        object_id: str,
        extensions: Optional[list[str]] = None,
        recursive: bool = False,
        previous_path: Optional[str] = None,
    ) -> list[dict]:
        fields_input = "nextPageToken, files({})".format(",".join(self.fields))
        q = f"'{object_id}' in parents"
        # Filter by extension but still include any directories
        if extensions:
            ext_filter = " or ".join([f"fileExtension = '{e}'" for e in extensions])
            q = f"{q} and ({ext_filter} or mimeType = 'application/vnd.google-apps.folder')"
        logger.debug(f"query used when indexing: {q}")
        logger.debug("response fields limited to: {}".format(", ".join(self.fields)))
        done = False
        page_token = None
        files_response = []
        while not done:
            response: dict = files_client.list(
                supportsAllDrives=True,
                includeItemsFromAllDrives=True,
                spaces="drive",
                fields=fields_input,
                corpora="user",
                pageToken=page_token,
                q=q,
            ).execute()
            if files := response.get("files", []):
                fs = [f for f in files if not self.is_dir(record=f)]
                for r in fs:
                    r["parent_path"] = previous_path
                dirs = [f for f in files if self.is_dir(record=f)]
                files_response.extend(fs)
                if recursive:
                    for d in dirs:
                        dir_id = d["id"]
                        dir_name = d["name"]
                        files_response.extend(
                            self.get_paginated_results(
                                files_client=files_client,
                                object_id=dir_id,
                                extensions=extensions,
                                recursive=recursive,
                                previous_path=f"{previous_path}/{dir_name}",
                            )
                        )
            page_token = response.get("nextPageToken")
            if page_token is None:
                done = True
        for r in files_response:
            r["parent_root_path"] = previous_path
        return files_response

    def get_root_info(self, files_client, object_id: str) -> dict:
        return files_client.get(
            supportsAllDrives=True, fileId=object_id, fields=",".join(self.fields)
        ).execute()

    def get_files(
        self,
        files_client,
        object_id: str,
        recursive: bool = False,
        extensions: Optional[list[str]] = None,
    ) -> list[FileData]:
        root_info = self.get_root_info(files_client=files_client, object_id=object_id)
        if not self.is_dir(root_info):
            root_info["permissions"] = self.extract_permissions(root_info.get("permissions"))
            data = [self.map_file_data(root_info)]
        else:
            file_contents = self.get_paginated_results(
                files_client=files_client,
                object_id=object_id,
                extensions=extensions,
                recursive=recursive,
                previous_path=root_info["name"],
            )
            data = []
            for f in file_contents:
                f["permissions"] = self.extract_permissions(f.get("permissions"))
                data.append(self.map_file_data(root_info=f))
        for d in data:
            d.metadata.record_locator["drive_id"]: object_id
        return data

    def extract_permissions(self, permissions: Optional[list[dict]]) -> list[dict]:
        if not permissions:
            logger.debug("no permissions found")
            return [{}]

        # https://developers.google.com/workspace/drive/api/guides/ref-roles
        role_mapping = {
            "owner": ["read", "update", "delete"],
            "organizer": ["read", "update", "delete"],
            "fileOrganizer": ["read", "update"],
            "writer": ["read", "update"],
            "commenter": ["read"],
            "reader": ["read"],
        }

        normalized_permissions = {
            "read": {"users": set(), "groups": set()},
            "update": {"users": set(), "groups": set()},
            "delete": {"users": set(), "groups": set()},
        }

        for item in permissions:
            # https://developers.google.com/workspace/drive/api/reference/rest/v3/permissions
            # ignore permissions for "anyone" and "domain"
            if item["type"] in ["user", "group"]:
                type_key = item["type"] + "s"
                for operation in role_mapping[item["role"]]:
                    normalized_permissions[operation][type_key].add(item["id"])

        # turn sets into sorted lists for consistency and json serialization
        for role_dict in normalized_permissions.values():
            for key in role_dict:
                role_dict[key] = sorted(role_dict[key])

        logger.debug(f"normalized permissions generated: {normalized_permissions}")
        return [{k: v} for k, v in normalized_permissions.items()]

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        with self.connection_config.get_client() as client:
            for f in self.get_files(
                files_client=client,
                object_id=self.connection_config.drive_id,
                recursive=self.index_config.recursive,
                extensions=self.index_config.extensions,
            ):
                yield f


class GoogleDriveDownloaderConfig(DownloaderConfig):
    lro_max_tries: int = 10
    lro_max_time: int = 10 * 60  # 10 minutes


def _get_extension(file_data: FileData) -> str:
    """
    Returns the extension for a given source MIME type.
    """
    source_mime_type = file_data.additional_metadata.get("export_mime_type", "")
    export_mime_type = GOOGLE_EXPORT_MIME_MAP.get(source_mime_type, "")
    if export_mime_type:
        return EXPORT_EXTENSION_MAP.get(export_mime_type, "")
    return ""


@dataclass
class GoogleDriveDownloader(Downloader):
    """
    Downloads files from Google Drive using googleapis client. For native files, it uses the export
    functionality for files <10MB and LRO (Long Running Operation) for files >10MB.
    """

    connection_config: GoogleDriveConnectionConfig
    download_config: GoogleDriveDownloaderConfig = field(
        default_factory=lambda: GoogleDriveDownloaderConfig()
    )
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["googleapiclient"], extras="google-drive")
    def _direct_download_file(self, file_id, download_path: Path):
        """Downloads a file from Google Drive using the Drive API's media download functionality.
        The method uses Google Drive API's media download functionality to stream the file
        content directly to disk.

        Args:
            file_id (str): The ID of the file to download from Google Drive.
            download_path (Path): The local path where the file should be saved.

        Raises:
            SourceConnectionError: If the download operation fails.
        """
        from googleapiclient.errors import HttpError
        from googleapiclient.http import MediaIoBaseDownload

        try:
            with self.connection_config.get_client() as client:
                # pylint: disable=maybe-no-member
                request = client.get_media(fileId=file_id)

                with open(download_path, "wb") as file:
                    downloader = MediaIoBaseDownload(file, request)
                    done = False
                    while done is False:
                        status, done = downloader.next_chunk()
                        logger.debug(f"Download progress:{int(status.progress() * 100)}.")

        except (HttpError, ValueError) as error:
            logger.exception(f"Error downloading file {file_id} to {download_path}: {error}")
            raise SourceConnectionError("Failed to download file") from error

    @requires_dependencies(["googleapiclient"], extras="google-drive")
    def _export_gdrive_file_with_lro(self, file_id: str, download_path: Path, mime_type: str):
        """Exports a Google Drive file using Long-Running Operation (LRO) for large files
        (>10MB of the exported file size).

        This method is used when the standard export method fails due to file size limitations.
        It uses the Drive API's LRO functionality to handle large file exports.

        Args:
            file_id (str): The ID of the Google Drive file to export.
            download_path (Path): The local path where the exported file should be saved.
            mime_type (str): The target MIME type for the exported file.
        Raises:
            SourceConnectionError: If the export operation fails.
        """

        import tenacity
        from googleapiclient.errors import HttpError

        max_time = self.download_config.lro_max_time
        max_tries = self.download_config.lro_max_tries

        class OperationNotFinished(Exception):
            """
            Exception raised when the operation is not finished.
            """

            pass

        def is_fatal_code(e: Exception) -> bool:
            """
            Returns True if the error is fatal and should not be retried.
            403 and 429 can mean "Too many requests" or "User rate limit exceeded"
            which should be retried.
            """
            return (
                isinstance(e, HttpError)
                and 400 <= e.resp.status < 500
                and e.resp.status not in [403, 429]
            )

        @tenacity.retry(
            wait=tenacity.wait_exponential(),
            retry=tenacity.retry_if_exception(
                lambda e: (
                    isinstance(e, (HttpError, OperationNotFinished)) and not is_fatal_code(e)
                )
            ),
            stop=(tenacity.stop_after_attempt(max_tries) | tenacity.stop_after_delay(max_time)),
        )
        def _poll_operation(operation: dict, operations_client: "GoogleAPIResource") -> dict:
            """
            Helper function to poll the operation until it's complete.
            Uses backoff exponential retry logic.

            Each `operations.get` call uses the Google API requests limit. Details:
            https://developers.google.com/workspace/drive/api/guides/limits

            The limits as of May 2025 are:
            - 12.000 calls per 60 seconds

            In case of request limitting, the API will return 403 `User rate limit exceeded` error
            or 429 `Too many requests` error.
            """
            if operation.get("done", False):
                return operation
            if "error" in operation:
                raise SourceConnectionError(
                    f"Export operation failed: {operation['error']['message']}"
                )
            # Refresh the operation status:
            # FYI: In some cases the `operations.get` call errors with 403 "User does not have
            # permission" error even if the same user create the operation with `download` method.
            updated_operation = operations_client.get(name=operation["name"]).execute()
            if not updated_operation.get("done", False):
                raise OperationNotFinished()
            return updated_operation

        try:
            with self._get_files_and_operations_client() as (files_client, operations_client):
                # Start the LRO
                operation = files_client.download(fileId=file_id, mimeType=mime_type).execute()

                # In case the operation is not finished, poll it until it's complete
                updated_operation = _poll_operation(operation, operations_client)

                # Get the download URI from the completed operation
                download_uri = updated_operation["response"]["downloadUri"]

            # Download the file using the URI
            self._raw_download_google_drive_file(download_uri, download_path)

        except HttpError as error:
            raise SourceConnectionError(
                f"Failed to export file using Google Drive LRO: {error}"
            ) from error

    @requires_dependencies(["googleapiclient"], extras="google-drive")
    def _export_gdrive_native_file(
        self, file_id: str, download_path: Path, mime_type: str, file_size: int
    ):
        """Exports a Google Drive native file (Docs, Sheets, Slides) to a specified format.

        This method uses the Google Drive API's export functionality to convert Google Workspace
        files to other formats (e.g., Google Docs to PDF, Google Sheets to Excel).
        For files larger than 10MB, it falls back to using Long-Running Operation (LRO).

        Args:
            file_id (str): The ID of the Google Drive file to export.
            download_path (Path): The local path where the exported file should be saved.
            mime_type (str): The target MIME type for the exported file (e.g., 'application/pdf').
            file_size (int): The size of the file to export - used to determine if the
                file is large enough to use LRO instead of direct export endpoint.
        Returns:
            bytes: The exported file content.

        Raises:
            HttpError: If the export operation fails.
        """
        from googleapiclient.errors import HttpError
        from googleapiclient.http import MediaIoBaseDownload

        if file_size > LRO_EXPORT_SIZE_THRESHOLD:
            self._export_gdrive_file_with_lro(file_id, download_path, mime_type)
            return

        with self.connection_config.get_client() as client:
            try:
                # pylint: disable=maybe-no-member
                request = client.export_media(fileId=file_id, mimeType=mime_type)
                with open(download_path, "wb") as file:
                    downloader = MediaIoBaseDownload(file, request)
                    done = False
                    while done is False:
                        status, done = downloader.next_chunk()
                        logger.debug(f"Download progress: {int(status.progress() * 100)}.")
            except HttpError as error:
                if error.resp.status == 403 and "too large" in error.reason.lower():
                    # Even though we have the LRO threashold, for some smaller files the
                    # export size might exceed 10MB and we get a 403 error.
                    # In that case, we use LRO as a fallback.
                    self._export_gdrive_file_with_lro(file_id, download_path, mime_type)
                else:
                    raise SourceConnectionError(f"Failed to export file: {error}") from error

    @requires_dependencies(["googleapiclient"], extras="google-drive")
    @contextmanager
    def _get_files_and_operations_client(
        self,
    ) -> Generator[tuple["GoogleAPIResource", "GoogleAPIResource"], None, None]:
        """
        Returns a context manager for the files and operations clients for the Google Drive API.

        Yields:
            Tuple[GoogleAPIResource, GoogleAPIResource]: A tuple of the files
                and operations clients.
        """
        from googleapiclient.discovery import build

        creds = self._get_credentials()
        service = build("drive", "v3", credentials=creds)
        with (
            service.operations() as operations_client,
            service.files() as files_client,
        ):
            yield files_client, operations_client

    @requires_dependencies(["httpx"])
    def _raw_download_google_drive_file(self, url: str, download_path: Path) -> Path:
        """
        Streams file content directly to disk using authenticated HTTP request.
        Must use httpx to stream the file to disk as currently there's no google SDK
        functionality to download a file like for get media or export operations.

        Writes the file to the correct path in the download directory while downloading.
        Avoids buffering large files in memory.

        Args:
            url (str): The URL of the file to download.
            download_path (Path): The path to save the downloaded file.

        Returns:
            Path: The path to the downloaded file.
        """
        import httpx
        from google.auth.transport.requests import Request

        creds = self._get_credentials()

        creds.refresh(Request())

        headers = {
            "Authorization": f"Bearer {creds.token}",
        }

        with (
            httpx.Client(timeout=None, follow_redirects=True) as client,
            client.stream("GET", url, headers=headers) as response,
        ):
            if response.status_code != 200:
                raise SourceConnectionError(
                    f"Failed to stream download from {url}: {response.status_code}"
                )
            with open(download_path, "wb") as f:
                for chunk in response.iter_bytes():
                    f.write(chunk)
        return download_path

    @requires_dependencies(["google"], extras="google-drive")
    def _get_credentials(self):
        """
        Retrieves the credentials for Google Drive API access.

        Returns:
            Credentials: The credentials for Google Drive API access.
        """
        from google.oauth2 import service_account

        access_config = self.connection_config.access_config.get_secret_value()
        key_data = access_config.get_service_account_key()
        creds = service_account.Credentials.from_service_account_info(
            key_data,
            scopes=["https://www.googleapis.com/auth/drive.readonly"],
        )
        return creds

    def _download_file(self, file_data: FileData) -> Path:
        """Downloads a file from Google Drive using either direct download or export based
        on the source file's MIME type.

        This method determines the appropriate download method based on the file's MIME type:
        - For Google Workspace files (Docs, Sheets, Slides), uses export functionality
        - For other files, uses direct download

        Args:
            file_data (FileData): The metadata of the file being downloaded.

        Returns:
            Path: The path to the downloaded file.

        Raises:
            SourceConnectionError: If the download fails.
        """
        mime_type = file_data.additional_metadata.get("mimeType", "")
        file_size = int(file_data.additional_metadata.get("size", 0))
        file_id = file_data.identifier

        download_path = self.get_download_path(file_data)
        if not download_path:
            raise SourceConnectionError(f"Failed to get download path for file {file_id}")

        if mime_type in GOOGLE_EXPORT_MIME_MAP:
            # For Google Workspace files, use export functionality
            ext = _get_extension(file_data)
            download_path = download_path.with_suffix(ext)
            download_path.parent.mkdir(parents=True, exist_ok=True)
            export_mime = GOOGLE_EXPORT_MIME_MAP[mime_type]
            self._export_gdrive_native_file(
                file_id=file_id,
                download_path=download_path,
                mime_type=export_mime,
                file_size=file_size,
            )
            file_data.additional_metadata.update(
                {
                    "export_mime_type": export_mime,
                    "export_extension": ext,
                    "download_method": "google_workspace_export",
                }
            )
        else:
            # For other files, use direct download
            download_path.parent.mkdir(parents=True, exist_ok=True)
            self._direct_download_file(file_id=file_id, download_path=download_path)
            file_data.additional_metadata.update(
                {
                    "download_method": "direct_download",
                }
            )

        return download_path

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        mime_type = file_data.additional_metadata.get("mimeType", "")

        logger.debug(
            f"Downloading file {file_data.source_identifiers.fullpath} of type {mime_type}"
        )

        download_path = self._download_file(file_data)

        file_data.local_download_path = str(download_path.resolve())

        return self.generate_download_response(file_data=file_data, download_path=download_path)


google_drive_source_entry = SourceRegistryEntry(
    connection_config=GoogleDriveConnectionConfig,
    indexer_config=GoogleDriveIndexerConfig,
    indexer=GoogleDriveIndexer,
    downloader_config=GoogleDriveDownloaderConfig,
    downloader=GoogleDriveDownloader,
)



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pydantic import Field, SecretStr

from unstructured_ingest.embed.interfaces import (
    AsyncBaseEmbeddingEncoder,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.error import (
    ProviderError,
    UserAuthError,
    UserError,
    is_internal_error,
)
from unstructured_ingest.error import (
    RateLimitError as CustomRateLimitError,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from together import AsyncTogether, Together


class TogetherAIEmbeddingConfig(EmbeddingConfig):
    api_key: SecretStr = Field(description="API key for Together AI")
    embedder_model_name: str = Field(
        default="togethercomputer/m2-bert-80M-32k-retrieval",
        alias="model_name",
        description="Together AI model name",
    )

    def wrap_error(self, e: Exception) -> Exception:
        if is_internal_error(e=e):
            return e
        # https://docs.together.ai/docs/error-codes
        from together.error import AuthenticationError, RateLimitError, TogetherException

        if not isinstance(e, TogetherException):
            logger.error(f"unhandled exception from together: {e}", exc_info=True)
            return e
        message = e.args[0]
        if isinstance(e, AuthenticationError):
            return UserAuthError(message)
        if isinstance(e, RateLimitError):
            return CustomRateLimitError(message)

        status_code = getattr(e, "status_code", None)
        if status_code is not None:
            if 400 <= status_code < 500:
                return UserError(message)
            if status_code >= 500:
                return ProviderError(message)
        logger.error(f"unhandled exception from together: {e}", exc_info=True)
        return e

    def run_precheck(self) -> None:
        client = self.get_client()
        try:
            models = [m.id for m in list(client.models.list())]
            if self.embedder_model_name not in models:
                raise UserError(
                    "model '{}' not found: {}".format(self.embedder_model_name, ", ".join(models))
                )
        except Exception as e:
            raise self.wrap_error(e=e)

    @requires_dependencies(["together"], extras="togetherai")
    def get_client(self) -> "Together":
        from together import Together

        return Together(api_key=self.api_key.get_secret_value())

    @requires_dependencies(["together"], extras="togetherai")
    def get_async_client(self) -> "AsyncTogether":
        from together import AsyncTogether

        return AsyncTogether(api_key=self.api_key.get_secret_value())


@dataclass
class TogetherAIEmbeddingEncoder(BaseEmbeddingEncoder):
    config: TogetherAIEmbeddingConfig

    def precheck(self):
        self.config.run_precheck()

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def get_client(self) -> "Together":
        return self.config.get_client()

    def embed_batch(self, client: "Together", batch: list[str]) -> list[list[float]]:
        outputs = client.embeddings.create(model=self.config.embedder_model_name, input=batch)
        return [outputs.data[i].embedding for i in range(len(batch))]


@dataclass
class AsyncTogetherAIEmbeddingEncoder(AsyncBaseEmbeddingEncoder):
    config: TogetherAIEmbeddingConfig

    def precheck(self):
        self.config.run_precheck()

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def get_client(self) -> "AsyncTogether":
        return self.config.get_async_client()

    async def embed_batch(self, client: Any, batch: list[str]) -> list[list[float]]:
        outputs = await client.embeddings.create(model=self.config.embedder_model_name, input=batch)
        return [outputs.data[i].embedding for i in range(len(batch))]



import json
import re
from datetime import datetime
from typing import Any, Union

from dateutil import parser

from unstructured_ingest.logger import logger


def json_to_dict(json_string: str) -> Union[str, dict[str, Any]]:
    """Helper function attempts to deserialize json string to a dictionary."""
    try:
        return json.loads(json_string)
    except json.JSONDecodeError:
        # Not necessary an error if it is a path or malformed json
        pass
    try:
        # This is common when single quotes are used instead of double quotes
        return json.loads(json_string.replace("'", '"'))
    except json.JSONDecodeError:
        # Not necessary an error if it is a path
        pass
    return json_string


def ensure_isoformat_datetime(timestamp: Union[datetime, str]) -> str:
    """
    Ensures that the input value is converted to an ISO format datetime string.
    Handles both datetime objects and strings.
    """
    if isinstance(timestamp, datetime):
        return timestamp.isoformat()
    elif isinstance(timestamp, str):
        try:
            # Parse the datetime string in various formats
            dt = parser.parse(timestamp)
            return dt.isoformat()
        except ValueError as e:
            raise ValueError(f"String '{timestamp}' could not be parsed as a datetime.") from e
    else:
        raise TypeError(f"Expected input type datetime or str, but got {type(timestamp)}.")


def truncate_string_bytes(string: str, max_bytes: int, encoding: str = "utf-8") -> str:
    """
    Truncates a string to a specified maximum number of bytes.
    """
    encoded_string = str(string).encode(encoding)
    if len(encoded_string) <= max_bytes:
        return string
    return encoded_string[:max_bytes].decode(encoding, errors="ignore")


def fix_unescaped_unicode(text: str, encoding: str = "utf-8") -> str:
    """
    Fix unescaped Unicode sequences in text.
    """
    try:
        _text: str = json.dumps(text)

        # Pattern to match unescaped Unicode sequences like \\uXXXX
        pattern = r"\\\\u([0-9A-Fa-f]{4})"
        # Replace with properly escaped Unicode sequences \uXXXX
        _text = re.sub(pattern, r"\\u\1", _text)
        _text = json.loads(_text)

        # Encode the text to check for encoding errors
        _text.encode(encoding)
        return _text
    except Exception as e:
        # Return original text if encoding fails
        logger.warning(f"Failed to fix unescaped Unicode sequences: {e}", exc_info=True)
        return text



# https://developers.notion.com/reference/property-object#date
from dataclasses import dataclass, field
from typing import Optional

from htmlBuilder.tags import HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import DBCellBase, DBPropertyBase
from unstructured_ingest.processes.connectors.notion.types.date import Date as DateType


@dataclass
class Date(DBPropertyBase):
    id: str
    name: str
    type: str = "date"
    description: Optional[str] = None
    date: dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class DateCell(DBCellBase):
    id: str
    date: Optional[DateType] = None
    name: Optional[str] = None
    type: str = "date"

    @classmethod
    def from_dict(cls, data: dict):
        date = None
        date_data = data.pop("date")
        if date_data:
            date = DateType.from_dict(date_data)
        return cls(date=date, **data)

    def get_html(self) -> Optional[HtmlTag]:
        if date := self.date:
            return date.get_html()
        return None



from pathlib import Path
from typing import Any, Dict, Optional, Union
from urllib.parse import urlparse


class DataSanitizer:
    """Utility class for sanitizing sensitive data in logs."""

    @staticmethod
    def sanitize_path(path: Union[str, Path]) -> str:
        """Sanitize file paths for logging, showing only filename and partial path."""
        if not path:
            return "<empty>"

        path_str = str(path)
        path_obj = Path(path_str)

        if len(path_obj.parts) > 2:
            return f".../{path_obj.parent.name}/{path_obj.name}"
        return path_obj.name

    @staticmethod
    def sanitize_id(identifier: str) -> str:
        """Sanitize IDs for logging, showing only first/last few characters."""
        if not identifier:
            return "<id>"
        if len(identifier) < 10:
            half_len = len(identifier) // 2
            return f"{identifier[:half_len]}..."
        return f"{identifier[:4]}...{identifier[-4:]}"

    @staticmethod
    def sanitize_url(url: str) -> str:
        """Sanitize URLs for logging, removing sensitive query parameters."""
        if not url:
            return "<url>"
        try:
            parsed = urlparse(url)
            return f"{parsed.scheme}://{parsed.netloc}{parsed.path}"
        except (ValueError, TypeError):
            return "<url>"

    @staticmethod
    def sanitize_token(token: str) -> str:
        """Sanitize tokens and secrets for logging."""
        if not token:
            return "<token>"
        if len(token) < 10:
            half_len = len(token) // 2
            return f"{token[:half_len]}..."
        return f"{token[:4]}...{token[-4:]}"

    @staticmethod
    def sanitize_location(location: Union[str, Path]) -> str:
        """Sanitize document locations (file paths, URLs, database references) for logging."""
        if not location:
            return "<empty>"

        location_str = str(location)

        # Handle URLs
        if location_str.startswith(("http://", "https://", "ftp://", "ftps://")):
            return DataSanitizer.sanitize_url(location_str)

        # Handle database-style references (table:id, collection/document, etc.)
        if ":" in location_str and not location_str.startswith("/"):
            parts = location_str.split(":", 1)
            if len(parts) == 2:
                table_name, record_id = parts
                return f"{table_name}:{DataSanitizer.sanitize_id(record_id)}"

        return DataSanitizer.sanitize_path(location_str)

    @staticmethod
    def sanitize_document_id(document_id: str) -> str:
        """Sanitize document IDs for logging (alias for sanitize_id for clarity)."""
        return DataSanitizer.sanitize_id(document_id)

    @staticmethod
    def sanitize_dict(data: Dict[str, Any], sensitive_keys: Optional[set] = None) -> Dict[str, Any]:
        """Sanitize dictionary data for logging."""
        if sensitive_keys is None:
            sensitive_keys = {
                "password",
                "token",
                "secret",
                "key",
                "api_key",
                "access_token",
                "refresh_token",
                "client_secret",
                "private_key",
                "credentials",
            }

        sanitized = {}
        for k, v in data.items():
            key_lower = k.lower()
            if any(sensitive_key in key_lower for sensitive_key in sensitive_keys):
                sanitized[k] = DataSanitizer.sanitize_token(str(v))
            elif isinstance(v, dict):
                sanitized[k] = DataSanitizer.sanitize_dict(v, sensitive_keys)
            elif isinstance(v, (str, Path)) and (
                "path" in key_lower
                or "file" in key_lower
                or "location" in key_lower
                or "document_location" in key_lower
            ):
                sanitized[k] = DataSanitizer.sanitize_location(v)
            elif isinstance(v, str) and (
                ("id" in key_lower and len(str(v)) > 8)
                or ("document_id" in key_lower and len(str(v)) > 8)
            ):
                sanitized[k] = DataSanitizer.sanitize_document_id(v)
            else:
                sanitized[k] = v
        return sanitized



# https://developers.notion.com/reference/block#child-page
from dataclasses import dataclass
from typing import Optional

from htmlBuilder.tags import HtmlTag, P

from unstructured_ingest.processes.connectors.notion.interfaces import BlockBase, GetHTMLMixin


@dataclass
class ChildPage(BlockBase, GetHTMLMixin):
    title: str

    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        return P([], self.title)



# https://developers.notion.com/reference/property-object#title
from dataclasses import dataclass, field
from typing import Optional

from htmlBuilder.tags import Div, HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import (
    DBCellBase,
    DBPropertyBase,
    FromJSONMixin,
)


@dataclass
class UniqueID(DBPropertyBase):
    id: str
    name: str
    type: str = "unique_id"
    unique_id: dict = field(default_factory=dict)
    description: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class UniqueIDCellData(FromJSONMixin):
    prefix: str
    number: int

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class UniqueIDCell(DBCellBase):
    id: str
    unique_id: Optional[UniqueIDCellData]
    type: str = "title"
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(unique_id=UniqueIDCellData.from_dict(data.pop("unique_id")), **data)

    def get_html(self) -> Optional[HtmlTag]:
        if unique_id := self.unique_id:
            return Div([], f"{unique_id.prefix}-{unique_id.number}")
        return None



from .connector import AccessConfig, BaseConnector, ConnectionConfig
from .downloader import Downloader, DownloaderConfig, DownloadResponse, download_responses
from .indexer import Indexer, IndexerConfig
from .process import BaseProcess
from .processor import ProcessorConfig
from .upload_stager import UploadStager, UploadStagerConfig
from .uploader import UploadContent, Uploader, UploaderConfig, VectorDBUploader

__all__ = [
    "DownloadResponse",
    "download_responses",
    "Downloader",
    "DownloaderConfig",
    "Indexer",
    "IndexerConfig",
    "BaseProcess",
    "ProcessorConfig",
    "UploadStager",
    "UploadStagerConfig",
    "Uploader",
    "UploaderConfig",
    "UploadContent",
    "AccessConfig",
    "ConnectionConfig",
    "BaseConnector",
    "VectorDBUploader",
]



# https://developers.notion.com/reference/block#synced-block
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.tags import HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import BlockBase


@dataclass
class OriginalSyncedBlock(BlockBase):
    synced_from: Optional[str] = None
    children: List[dict] = field(default_factory=list)

    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        """Create OriginalSyncedBlock from dictionary data.

        Original blocks contain children content.
        """
        if "children" not in data:
            raise ValueError(f"OriginalSyncedBlock data missing 'children': {data}")
        return cls(children=data["children"])

    def get_html(self) -> Optional[HtmlTag]:
        return None


@dataclass
class DuplicateSyncedBlock(BlockBase):
    type: str
    block_id: str

    @staticmethod
    def can_have_children() -> bool:
        """Check if duplicate synced blocks can have children.

        Duplicate blocks themselves don't have children directly fetched here,
        but they represent content that does, so Notion API might report has_children=True
        on the parent block object. The actual children are fetched from the original block.
        """
        return True

    @classmethod
    def from_dict(cls, data: dict):
        """Create DuplicateSyncedBlock from dictionary data.

        Duplicate blocks contain a 'synced_from' reference.
        """
        synced_from_data = data.get("synced_from")
        if not synced_from_data or not isinstance(synced_from_data, dict):
            raise ValueError(f"Invalid data structure for DuplicateSyncedBlock: {data}")
        # Ensure required keys are present in the nested dictionary
        if "type" not in synced_from_data or "block_id" not in synced_from_data:
            raise ValueError(
                f"Missing 'type' or 'block_id' in synced_from data: {synced_from_data}"
            )
        return cls(type=synced_from_data["type"], block_id=synced_from_data["block_id"])

    def get_html(self) -> Optional[HtmlTag]:
        """Get HTML representation of the duplicate synced block.

        HTML representation might need fetching the original block's content,
        which is outside the scope of this simple data class.
        """
        return None


class SyncBlock(BlockBase):
    @staticmethod
    def can_have_children() -> bool:
        """Check if synced blocks can have children.

        Synced blocks (both original and duplicate) can conceptually have children.
        """
        return True

    @classmethod
    def from_dict(cls, data: dict):
        """Create appropriate SyncedBlock subclass from dictionary data.

        Determine if it's a duplicate (has 'synced_from') or original (has 'children').
        """
        if data.get("synced_from") is not None:
            # It's a duplicate block containing a reference
            return DuplicateSyncedBlock.from_dict(data)
        elif "children" in data:
            # It's an original block containing children
            return OriginalSyncedBlock.from_dict(data)
        else:
            # Handle cases where neither 'synced_from' nor 'children' are present.
            # Notion API might return this for an empty original synced block.
            # Let's treat it as an empty OriginalSyncedBlock.
            # If this assumption is wrong, errors might occur later.
            # Consider logging a warning here if strictness is needed.
            return OriginalSyncedBlock(children=[])

    def get_html(self) -> Optional[HtmlTag]:
        """Get HTML representation of the synced block.

        The specific instance returned by from_dict (Original or Duplicate)
        will handle its own get_html logic.
        This method on the base SyncBlock might not be directly called.
        """
        return None



# https://developers.notion.com/reference/block#link-to-page
from dataclasses import dataclass
from typing import Optional

from htmlBuilder.tags import Div, HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import BlockBase


@dataclass
class LinkToPage(BlockBase):
    type: str
    page_id: Optional[str] = None
    database_id: Optional[str] = None

    @staticmethod
    def can_have_children() -> bool:
        return False

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        if page_id := self.page_id:
            return Div([], page_id)
        if database_id := self.database_id:
            return Div([], database_id)
        return None



import logging
import traceback
from dataclasses import dataclass, field
from multiprocessing import Queue, current_process
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional
from urllib.parse import urlparse

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.error import DestinationConnectionError, ValueError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.utils.constants import RECORD_ID_LABEL
from unstructured_ingest.utils.data_prep import get_data_df, get_json_data
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.table import convert_to_pandas_dataframe

CONNECTOR_TYPE = "delta_table"

if TYPE_CHECKING:
    from pandas import DataFrame


@requires_dependencies(["deltalake"], extras="delta-table")
def write_deltalake_with_error_handling(queue, **kwargs):
    from deltalake.writer import write_deltalake

    try:
        write_deltalake(**kwargs)
    except Exception:
        queue.put(traceback.format_exc())


class DeltaTableAccessConfig(AccessConfig):
    aws_access_key_id: Optional[str] = Field(default=None, description="AWS Access Key Id")
    aws_secret_access_key: Optional[str] = Field(default=None, description="AWS Secret Access Key")


class DeltaTableConnectionConfig(ConnectionConfig):
    access_config: Secret[DeltaTableAccessConfig] = Field(
        default_factory=lambda: Secret[DeltaTableAccessConfig](DeltaTableAccessConfig()),
        validate_default=True,
    )
    aws_region: Optional[str] = Field(default=None, description="AWS Region")
    table_uri: str = Field(
        description=(
            "Local path or path to the target folder in the S3 bucket, "
            "formatted as s3://my-bucket/my-folder/"
        ),
    )

    def update_storage_options(self, storage_options: dict[str, str]) -> None:
        secrets = self.access_config.get_secret_value()
        if self.aws_region and secrets.aws_access_key_id and secrets.aws_secret_access_key:
            storage_options["AWS_REGION"] = self.aws_region
            storage_options["AWS_ACCESS_KEY_ID"] = secrets.aws_access_key_id
            storage_options["AWS_SECRET_ACCESS_KEY"] = secrets.aws_secret_access_key
            # Delta-rs doesn't support concurrent S3 writes without external locks (DynamoDB).
            # This flag allows single-writer uploads to S3 without using locks, according to:
            # https://delta-io.github.io/delta-rs/usage/writing/writing-to-s3-with-locking-provider/
            storage_options["AWS_S3_ALLOW_UNSAFE_RENAME"] = "true"


class DeltaTableUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class DeltaTableUploadStager(UploadStager):
    upload_stager_config: DeltaTableUploadStagerConfig = field(
        default_factory=lambda: DeltaTableUploadStagerConfig()
    )

    def run(  # type: ignore[override]
        self,
        elements_filepath: Path,
        file_data: FileData,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        elements_contents = get_json_data(elements_filepath)
        output_path = Path(output_dir) / Path(f"{output_filename}.parquet")

        df = convert_to_pandas_dataframe(elements_dict=elements_contents)
        # Ensure per-record overwrite/delete semantics: tag each row with the record identifier
        df[RECORD_ID_LABEL] = file_data.identifier
        df = df.dropna(axis=1, how="all")
        df.to_parquet(output_path)

        return output_path


class DeltaTableUploaderConfig(UploaderConfig):
    pass


@dataclass
class DeltaTableUploader(Uploader):
    upload_config: DeltaTableUploaderConfig
    connection_config: DeltaTableConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["boto3"], extras="delta-table")
    def precheck(self):
        secrets = self.connection_config.access_config.get_secret_value()
        if (
            self.connection_config.aws_region
            and secrets.aws_access_key_id
            and secrets.aws_secret_access_key
        ):
            from boto3 import client

            url = urlparse(self.connection_config.table_uri)
            bucket_name = url.netloc
            dir_path = url.path.lstrip("/")

            try:
                s3_client = client(
                    "s3",
                    aws_access_key_id=secrets.aws_access_key_id,
                    aws_secret_access_key=secrets.aws_secret_access_key,
                )
                s3_client.put_object(Bucket=bucket_name, Key=dir_path, Body=b"")

                response = s3_client.get_bucket_location(Bucket=bucket_name)

                bucket_region = _normalize_location_constraint(response.get("LocationConstraint"))

                if self.connection_config.aws_region != bucket_region:
                    raise ValueError(
                        "Wrong AWS region provided: bucket "
                        f"'{bucket_name}' resides in '{bucket_region}', "
                        "but configuration specifies "
                        f"'{self.connection_config.aws_region}'."
                    )

            except Exception as e:
                logger.error(f"failed to validate connection: {e}", exc_info=True)
                raise DestinationConnectionError(f"failed to validate connection: {e}")

    @requires_dependencies(["tenacity"], extras="delta-table")
    def upload_dataframe(self, df: "DataFrame", file_data: FileData) -> None:
        upload_path = self.connection_config.table_uri
        logger.info(
            f"writing {len(df)} rows to destination table at {upload_path}\ndtypes: {df.dtypes}",
        )
        storage_options: dict[str, str] = {}
        self.connection_config.update_storage_options(storage_options=storage_options)

        # Decide whether the Delta table already exists. If it does, we first delete all rows
        # belonging to the current record and then append the fresh data. Otherwise we will
        # create a brand-new table via an overwrite.

        mode = "overwrite"
        try:
            from deltalake import DeltaTable  # pylint: disable=import-error

            dt = DeltaTable(upload_path, storage_options=storage_options)
            logger.debug(f"Table exists: deleting rows for {file_data.identifier}")
            # Table exists – remove any previous rows for this record_id so that appending is
            # effectively an idempotent overwrite for the record.
            dt.delete(predicate=f"{RECORD_ID_LABEL} = '{file_data.identifier}'")
            mode = "append"
        except Exception:
            # Table does not exist yet (or cannot be opened) – we will create it below with
            # mode="overwrite". All other failures will be captured later by the writer.
            logger.debug("Table does not exist: creating new table")

        writer_kwargs = {
            "table_or_uri": upload_path,
            "data": df,
            "mode": mode,
            "schema_mode": "merge",
            "storage_options": storage_options,
        }

        from tenacity import (
            before_log,
            retry,
            retry_if_exception,
            stop_after_attempt,
            wait_random,
        )

        def _is_commit_conflict(exc: BaseException) -> bool:  # noqa: ANN401
            """Return True if exception looks like a Delta Lake commit conflict.

            Besides the canonical *CommitFailed* / *Metadata changed* errors that
            deltalake surfaces when two writers clash, we occasionally hit
            messages such as *Delta transaction failed, version 0 already
            exists* while multiple processes race to create the very first log
            entry. These situations are equally safe to retry, so detect them
            too.
            """

            return isinstance(exc, RuntimeError) and any(
                marker in str(exc)
                for marker in (
                    "CommitFailed",
                    "Metadata changed",
                    "version 0 already exists",
                    "version already exists",
                    "Delta transaction failed",
                )
            )

        @retry(
            stop=stop_after_attempt(10),
            wait=wait_random(min=0.2, max=1.0),
            before=before_log(logger, logging.DEBUG),
            retry=retry_if_exception(_is_commit_conflict),
            reraise=True,
        )
        def _single_attempt() -> None:
            """One optimistic transaction: delete old rows, then append new ones."""

            # NOTE: deltalake writer on Linux sometimes can finish but still trigger a SIGABRT and
            # cause ingest to fail, even though all tasks are completed normally. Putting the writer
            # into a process mitigates this issue by ensuring python interpreter waits properly for
            # deltalake's rust backend to finish
            # Use a multiprocessing context that relies on 'spawn' to avoid inheriting the
            # parent process' Tokio runtime, which leads to `pyo3_runtime.PanicException`.
            from multiprocessing import get_context

            ctx = get_context("spawn")
            queue: "Queue[str]" = ctx.Queue()

            if current_process().daemon:
                # write_deltalake_with_error_handling will push any traceback to our queue
                write_deltalake_with_error_handling(queue=queue, **writer_kwargs)
            else:
                # On non-daemon processes we still guard against SIGABRT by running in a
                # dedicated subprocess created via the 'spawn' method.
                writer = ctx.Process(
                    target=write_deltalake_with_error_handling,
                    kwargs={"queue": queue, **writer_kwargs},
                )
                writer.start()
                writer.join()

                # First surface any traceback captured inside the subprocess so users see the real
                # root-cause instead of a generic non-zero exit code.
                if not queue.empty():
                    error_message = queue.get()
                    logger.error("Exception occurred in write_deltalake: %s", error_message)
                    raise DestinationConnectionError(f"Error in write_deltalake: {error_message}")

                # If the subprocess terminated abnormally but produced no traceback (e.g., SIGABRT),
                # still raise a helpful error for callers.
                if not current_process().daemon and writer.exitcode != 0:
                    raise DestinationConnectionError(
                        f"write_deltalake subprocess exited with code {writer.exitcode}"
                    )

        _single_attempt()

    @requires_dependencies(["pandas"], extras="delta-table")
    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        import pandas as pd

        df = pd.DataFrame(data=data)
        self.upload_dataframe(df=df, file_data=file_data)

    @requires_dependencies(["pandas"], extras="delta-table")
    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:  # type: ignore[override]
        df = get_data_df(path)
        self.upload_dataframe(df=df, file_data=file_data)


def _normalize_location_constraint(location: Optional[str]) -> str:
    """Return canonical AWS region name for a LocationConstraint value.

    The S3 GetBucketLocation operation returns `null` (`None`) for buckets in
    the legacy `us-east-1` region and `EU` for very old buckets that were
    created in the historical `EU` region (now `eu-west-1`). For every other
    region the API already returns the correct AWS region string. This helper
    normalises the legacy values so callers can reliably compare regions.

    Args:
        location: The LocationConstraint value returned by the S3 GetBucketLocation operation.

    Returns:
        The canonical AWS region name for the given location constraint.
    """

    if location is None:
        return "us-east-1"
    if location == "EU":
        return "eu-west-1"
    return location


delta_table_destination_entry = DestinationRegistryEntry(
    connection_config=DeltaTableConnectionConfig,
    uploader=DeltaTableUploader,
    uploader_config=DeltaTableUploaderConfig,
    upload_stager=DeltaTableUploadStager,
    upload_stager_config=DeltaTableUploadStagerConfig,
)



from functools import wraps
from typing import Callable, Optional

from unstructured_ingest.logger import logger
from unstructured_ingest.otel import OtelHandler


def instrument(
    span_name: Optional[str] = None,
    record_exception: bool = True,
    attributes: dict[str, str] = None,
    log_out: Callable = logger.info,
) -> Callable[[Callable], Callable]:
    def span_decorator(func: Callable) -> Callable:
        def get_name(self) -> str:
            if span_name:
                return span_name
            return f"{self.identifier} step"

        @wraps(func)
        def wrap_with_span(self, *args, **kwargs):
            name = get_name(self=self)
            otel_handler = OtelHandler(otel_endpoint=self.context.otel_endpoint, log_out=log_out)
            with otel_handler.get_tracer().start_as_current_span(
                name, record_exception=record_exception
            ) as span:
                otel_handler.set_attributes(span, attributes)
                return func(self, *args, **kwargs)

        return wrap_with_span

    return span_decorator



from collections import abc
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Callable, Generator, List, Optional, Union

from pydantic import BaseModel, Field, Secret

from unstructured_ingest.data_types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import SourceConnectionError, ValueError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
    download_responses,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    SourceRegistryEntry,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from atlassian import Jira

CONNECTOR_TYPE = "jira"

DEFAULT_C_SEP = " " * 5
DEFAULT_R_SEP = "\n"


class JiraIssueMetadata(BaseModel):
    id: str
    key: str
    fields: Optional[dict] = None  # Add fields to capture attachment data

    def get_project_id(self) -> str:
        return self.key.split("-")[0]

    def get_attachments(self) -> List[dict]:
        """Extract attachment information from fields"""
        if self.fields and "attachment" in self.fields:
            return self.fields["attachment"]
        return []


class FieldGetter(dict):
    def __getitem__(self, key):
        value = super().__getitem__(key) if key in self else None
        if value is None:
            value = FieldGetter({})
        return value


def nested_object_to_field_getter(obj: dict) -> Union[FieldGetter, dict]:
    if isinstance(obj, abc.Mapping):
        new_object = {}
        for k, v in obj.items():
            if isinstance(v, abc.Mapping):
                new_object[k] = FieldGetter(nested_object_to_field_getter(v))
            else:
                new_object[k] = v
        return FieldGetter(new_object)
    else:
        return obj


def api_token_based_generator(
    fn: Callable, key: str = "issues", **kwargs
) -> Generator[dict, None, None]:
    nextPageToken = kwargs.pop("nextPageToken", None)
    while True:
        resp = fn(nextPageToken=nextPageToken, **kwargs)
        issues = resp.get(key, [])
        for issue in issues:
            yield issue
        nextPageToken = resp.get("nextPageToken")
        if not nextPageToken:
            break


def api_page_based_generator(
    fn: Callable, key: str = "issues", **kwargs
) -> Generator[dict, None, None]:
    start = kwargs.pop("start", 0)
    while True:
        resp = fn(start=start, **kwargs)
        issues = resp.get(key, [])
        if not issues:
            break
        for issue in issues:
            yield issue
        start += len(issues)


class JiraAccessConfig(AccessConfig):
    password: Optional[str] = Field(
        description="Jira password or Cloud API token",
        default=None,
    )
    token: Optional[str] = Field(
        description="Jira Personal Access Token",
        default=None,
    )


class JiraConnectionConfig(ConnectionConfig):
    url: str = Field(description="URL of the Jira instance")
    username: Optional[str] = Field(
        description="Username or email for authentication",
        default=None,
    )
    cloud: bool = Field(description="Authenticate to Jira Cloud", default=False)
    access_config: Secret[JiraAccessConfig] = Field(description="Access configuration for Jira")

    def model_post_init(self, __context):
        access_configs = self.access_config.get_secret_value()
        basic_auth = self.username and access_configs.password
        pat_auth = access_configs.token
        if self.cloud and not basic_auth:
            raise ValueError(
                "cloud authentication requires username and API token (--password), "
                "see: https://atlassian-python-api.readthedocs.io/"
            )
        if basic_auth and pat_auth:
            raise ValueError(
                "both password and token provided, only one allowed, "
                "see: https://atlassian-python-api.readthedocs.io/"
            )
        if not (basic_auth or pat_auth):
            raise ValueError(
                "no form of auth provided, see: https://atlassian-python-api.readthedocs.io/"
            )

    @requires_dependencies(["atlassian"], extras="jira")
    @contextmanager
    def get_client(self) -> Generator["Jira", None, None]:
        from atlassian import Jira

        access_configs = self.access_config.get_secret_value()
        with Jira(
            url=self.url,
            username=self.username,
            password=access_configs.password,
            token=access_configs.token,
            cloud=self.cloud,
        ) as client:
            yield client


class JiraIndexerConfig(IndexerConfig):
    projects: Optional[list[str]] = Field(None, description="List of project keys")
    boards: Optional[list[str]] = Field(None, description="List of board IDs")
    issues: Optional[list[str]] = Field(None, description="List of issue keys or IDs")
    status_filters: Optional[list[str]] = Field(
        default=None,
        description="List of status filters, if provided will only return issues that have these statuses",  # noqa: E501
    )

    def model_post_init(self, context: Any, /) -> None:
        if not self.projects and not self.boards and not self.issues:
            raise ValueError("At least one of projects, boards, or issues must be provided.")


@dataclass
class JiraIndexer(Indexer):
    connection_config: JiraConnectionConfig
    index_config: JiraIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        try:
            with self.connection_config.get_client() as client:
                response = client.get_permissions("BROWSE_PROJECTS")
                permitted = response["permissions"]["BROWSE_PROJECTS"]["havePermission"]
        except Exception as e:
            logger.error(f"Failed to connect to Jira: {e}", exc_info=True)
            raise SourceConnectionError(f"Failed to connect to Jira: {e}")
        if not permitted:
            raise ValueError(
                """The provided user is not permitted to browse projects
                from the given Jira organization URL.
                Try checking username, password, token and the url arguments.""",
            )
        logger.info("Connection to Jira successful.")

    def run_jql(self, jql: str, **kwargs) -> Generator[JiraIssueMetadata, None, None]:
        with self.connection_config.get_client() as client:
            if client.cloud:
                for issue in api_token_based_generator(client.enhanced_jql, jql=jql, **kwargs):
                    yield JiraIssueMetadata.model_validate(issue)
            else:
                for issue in api_page_based_generator(client.jql, jql=jql, **kwargs):
                    yield JiraIssueMetadata.model_validate(issue)

    def _get_issues_within_projects(self) -> Generator[JiraIssueMetadata, None, None]:
        fields = ["key", "id", "status", "attachment"]  # Add attachment field
        jql = "project in ({})".format(", ".join(self.index_config.projects))
        jql = self._update_jql(jql)
        logger.debug(f"running jql: {jql}")
        return self.run_jql(jql=jql, fields=fields)

    def _get_issues_within_single_board(
        self, board_id: str
    ) -> Generator[JiraIssueMetadata, None, None]:
        with self.connection_config.get_client() as client:
            fields = ["key", "id", "attachment"]  # Add attachment field
            if self.index_config.status_filters:
                jql = "status in ({}) ORDER BY id".format(
                    ", ".join([f'"{s}"' for s in self.index_config.status_filters])
                )
            else:
                jql = "ORDER BY id"
            logger.debug(f"running jql for board {board_id}: {jql}")
            for issue in api_page_based_generator(
                fn=client.get_issues_for_board, board_id=board_id, fields=fields, jql=jql
            ):
                yield JiraIssueMetadata.model_validate(issue)

    def _get_issues_within_boards(self) -> Generator[JiraIssueMetadata, None, None]:
        if not self.index_config.boards:
            yield
        for board_id in self.index_config.boards:
            for issue in self._get_issues_within_single_board(board_id=board_id):
                yield issue

    def _update_jql(self, jql: str) -> str:
        if self.index_config.status_filters:
            jql += " and status in ({})".format(
                ", ".join([f'"{s}"' for s in self.index_config.status_filters])
            )
        jql = jql + " ORDER BY id"
        return jql

    def _get_issues_by_keys(self) -> Generator[JiraIssueMetadata, None, None]:
        fields = ["key", "id", "attachment"]  # Add attachment field
        jql = "key in ({})".format(", ".join(self.index_config.issues))
        jql = self._update_jql(jql)
        logger.debug(f"running jql: {jql}")
        return self.run_jql(jql=jql, fields=fields)

    def _create_file_data_from_issue(self, issue: JiraIssueMetadata) -> FileData:
        # Construct relative path and filename first
        filename = f"{issue.key}.txt"
        relative_path = str(Path(issue.get_project_id()) / filename)

        # Build metadata with attachments included in record_locator
        record_locator = {"id": issue.id, "key": issue.key, "full_path": relative_path}

        # Add attachments to record_locator if they exist
        attachments = issue.get_attachments()
        if attachments:
            record_locator["attachments"] = [
                {
                    "id": att["id"],
                    "filename": att["filename"],
                    "created": att.get("created"),
                    "mimeType": att.get("mimeType"),
                }
                for att in attachments
            ]

        metadata = FileDataSourceMetadata(
            date_processed=str(time()),
            record_locator=record_locator,
        )

        source_identifiers = SourceIdentifiers(
            filename=filename,
            fullpath=relative_path,
            rel_path=relative_path,
        )

        file_data = FileData(
            identifier=issue.id,
            connector_type=self.connector_type,
            metadata=metadata,
            additional_metadata=issue.model_dump(),
            source_identifiers=source_identifiers,
            display_name=source_identifiers.fullpath,
        )
        return file_data

    def get_generators(self) -> List[Callable]:
        generators = []
        if self.index_config.boards:
            generators.append(self._get_issues_within_boards)
        if self.index_config.issues:
            generators.append(self._get_issues_by_keys)
        if self.index_config.projects:
            generators.append(self._get_issues_within_projects)
        return generators

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        seen_keys = []
        for gen in self.get_generators():
            for issue in gen():
                if not issue:
                    continue
                if issue.key in seen_keys:
                    continue
                seen_keys.append(issue.key)
                yield self._create_file_data_from_issue(issue=issue)


class JiraDownloaderConfig(DownloaderConfig):
    download_attachments: bool = Field(
        default=False, description="If True, will download any attachments and process as well"
    )


@dataclass
class JiraDownloader(Downloader):
    connection_config: JiraConnectionConfig
    download_config: JiraDownloaderConfig = field(default_factory=JiraDownloaderConfig)
    connector_type: str = CONNECTOR_TYPE

    def _get_id_fields_for_issue(
        self, issue: dict, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        issue_id, key = issue["id"], issue["key"]
        return f"IssueID_IssueKey:{issue_id}{c_sep}{key}{r_sep}"

    def _get_project_fields_for_issue(
        self, issue: dict, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        if "project" in issue:
            return (
                f"ProjectID_Key:{issue['project']['key']}{c_sep}{issue['project']['name']}{r_sep}"
            )
        else:
            return ""

    def _get_dropdown_fields_for_issue(
        self, issue: dict, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        return f"""
        IssueType:{issue["issuetype"]["name"]}
        {r_sep}
        Status:{issue["status"]["name"]}
        {r_sep}
        Priority:{issue["priority"]}
        {r_sep}
        AssigneeID_Name:{issue["assignee"]["accountId"]}{c_sep}{issue["assignee"]["displayName"]}
        {r_sep}
        ReporterAdr_Name:{issue["reporter"]["emailAddress"]}{c_sep}{issue["reporter"]["displayName"]}
        {r_sep}
        Labels:{c_sep.join(issue["labels"])}
        {r_sep}
        Components:{c_sep.join([component["name"] for component in issue["components"]])}
        {r_sep}
        """

    def _get_subtasks_for_issue(self, issue: dict) -> str:
        return ""

    def _get_text_fields_for_issue(
        self, issue: dict, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        return f"""
        {issue["summary"]}
        {r_sep}
        {issue["description"]}
        {r_sep}
        {c_sep.join([attachment["self"] for attachment in issue["attachment"]])}
        {r_sep}
        """

    def _get_comments_for_issue(
        self, issue: dict, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        return c_sep.join(
            [self._get_fields_for_comment(comment) for comment in issue["comment"]["comments"]],
        )

    def _get_fields_for_comment(
        self, comment, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        return f"{comment['author']['displayName']}{c_sep}{comment['body']}{r_sep}"

    def form_templated_string(
        self,
        issue: dict,
        parsed_fields: Union[FieldGetter, dict],
        c_sep: str = "|||",
        r_sep: str = "\n\n\n",
    ) -> str:
        """Forms a template string via parsing the fields from the API response object on the issue
        The template string will be saved to the disk, and then will be processed by partition."""
        return r_sep.join(
            [
                self._get_id_fields_for_issue(issue),
                self._get_project_fields_for_issue(parsed_fields),
                self._get_dropdown_fields_for_issue(parsed_fields),
                self._get_subtasks_for_issue(parsed_fields),
                self._get_comments_for_issue(parsed_fields),
                self._get_text_fields_for_issue(parsed_fields),
            ],
        )

    def update_file_data(self, file_data: FileData, issue: dict) -> None:
        file_data.metadata.date_created = issue["fields"]["created"]
        file_data.metadata.date_modified = issue["fields"]["updated"]
        file_data.display_name = issue["fields"]["project"]["name"]

    def get_issue(self, issue_key: str) -> dict:
        try:
            with self.connection_config.get_client() as client:
                return client.issue(key=issue_key)
        except Exception as e:
            logger.error(f"Failed to fetch issue with key: {issue_key}: {e}", exc_info=True)
            raise SourceConnectionError(f"Failed to fetch issue with key: {issue_key}: {e}")

    def generate_attachment_file_data(
        self, attachment_dict: dict, parent_filedata: FileData
    ) -> FileData:
        new_filedata = parent_filedata.model_copy(deep=True)

        # Create attachment record_locator with parent context
        attachment_record_locator = {
            "id": attachment_dict["id"],
            "filename": attachment_dict["filename"],
            "created": attachment_dict.get("created"),
            "mimeType": attachment_dict.get("mimeType"),
            "parent": {
                "id": parent_filedata.metadata.record_locator["id"],
                "key": parent_filedata.metadata.record_locator["key"],
                "full_path": parent_filedata.source_identifiers.fullpath,
            },
        }

        # Append an identifier for attachment to not conflict with issue ids
        new_filedata.identifier = "{}a".format(attachment_dict["id"])
        filename = f"{attachment_dict['filename']}.{attachment_dict['id']}"
        new_filedata.metadata.filesize_bytes = attachment_dict.get("size")
        new_filedata.metadata.date_created = attachment_dict.get("created")
        new_filedata.metadata.url = attachment_dict.get("self")
        new_filedata.metadata.record_locator = attachment_record_locator
        full_path = (
            Path(parent_filedata.source_identifiers.fullpath).with_suffix("") / Path(filename)
        ).as_posix()
        new_filedata.metadata.record_locator["full_path"] = full_path
        new_filedata.source_identifiers = SourceIdentifiers(
            filename=filename,
            # add issue_parent to the fullpath and rel_path
            # to ensure that the attachment is saved in the same folder as the parent issue
            fullpath=full_path,
            rel_path=full_path,
        )
        return new_filedata

    def process_attachments(
        self, file_data: FileData, attachments: list[dict]
    ) -> list[DownloadResponse]:
        with self.connection_config.get_client() as client:
            download_path = self.get_download_path(file_data)
            attachment_download_dir = download_path.parent / "attachments"
            attachment_download_dir.mkdir(parents=True, exist_ok=True)
            download_responses = []
            for attachment in attachments:
                attachment_filename = Path(attachment["filename"])
                attachment_id = attachment["id"]
                attachment_download_path = attachment_download_dir / Path(
                    attachment_id
                ).with_suffix(attachment_filename.suffix)
                resp = client.get_attachment_content(attachment_id=attachment_id)
                with open(attachment_download_path, "wb") as f:
                    f.write(resp)
                attachment_filedata = self.generate_attachment_file_data(
                    attachment_dict=attachment, parent_filedata=file_data
                )
                download_responses.append(
                    self.generate_download_response(
                        file_data=attachment_filedata, download_path=attachment_download_path
                    )
                )
        return download_responses

    def run(self, file_data: FileData, **kwargs: Any) -> download_responses:
        issue_key = file_data.additional_metadata.get("key")
        if not issue_key:
            raise ValueError("Issue key not found in metadata.")
        issue = self.get_issue(issue_key)
        parsed_fields = nested_object_to_field_getter(issue["fields"])
        issue_str = self.form_templated_string(issue, parsed_fields)

        download_path = self.get_download_path(file_data)
        if download_path is None:
            raise ValueError("File data is missing source identifiers data.")
        download_path.parent.mkdir(parents=True, exist_ok=True)
        with open(download_path, "w") as f:
            f.write(issue_str)
        self.update_file_data(file_data, issue)
        download_response = self.generate_download_response(
            file_data=file_data, download_path=download_path
        )
        if self.download_config.download_attachments and (
            attachments := issue.get("fields", {}).get("attachment")
        ):
            attachment_responses = self.process_attachments(
                file_data=file_data, attachments=attachments
            )
            download_response = [download_response] + attachment_responses
        return download_response


jira_source_entry = SourceRegistryEntry(
    connection_config=JiraConnectionConfig,
    indexer_config=JiraIndexerConfig,
    indexer=JiraIndexer,
    downloader_config=JiraDownloaderConfig,
    downloader=JiraDownloader,
)



import enum
import logging
from dataclasses import dataclass, field
from typing import List, Optional, Tuple
from urllib.parse import urlparse
from uuid import UUID

from htmlBuilder.attributes import Style
from htmlBuilder.tags import (
    Body,
    Div,
    Head,
    Html,
    HtmlTag,
    Ol,
    Table,
    Td,
    Th,
    Title,
    Tr,
    Ul,
)
from notion_client.errors import APIResponseError

import unstructured_ingest.processes.connectors.notion.types.blocks as notion_blocks
from unstructured_ingest.processes.connectors.notion.client import Client
from unstructured_ingest.processes.connectors.notion.types.block import Block
from unstructured_ingest.processes.connectors.notion.types.database import Database


@dataclass
class HtmlExtractionResponse:
    html: Optional[HtmlTag] = None
    child_pages: List[str] = field(default_factory=list)
    child_databases: List[str] = field(default_factory=list)


def process_block(
    current_block: dict,
    parent_page_id: str,
    client: Client,
    child_pages: list,
    child_databases: list,
) -> Tuple[dict, list, list, dict]:
    if isinstance(current_block["block"].block, notion_blocks.ChildPage) and current_block[
        "block"
    ].id != str(parent_page_id):
        child_pages.append(current_block["block"].id)
        return {}, child_pages, child_databases
    if isinstance(current_block["block"].block, notion_blocks.ChildDatabase):
        child_databases.append(current_block["block"].id)
        return {}, child_pages, child_databases

    # recursively go through all blocks in a page, store each block in a dictionary
    if current_block["block"].has_children:
        children = []
        for children_block in client.blocks.children.iterate_list(
            block_id=current_block["block"].id
        ):
            children.extend(children_block)
        if children:
            for child in children:
                child_block = {
                    "block": child,
                    "level": current_block["level"] + 1,
                    "children": [],
                    "parent_id": current_block["block"].id,
                }
                child_element, child_pages, child_databases = process_block(
                    child_block, parent_page_id, client, child_pages, child_databases
                )
                current_block["children"].append(child_element)
    return current_block, child_pages, child_databases


def flush_list(type: str, item_list: list, html: list) -> Tuple[list, list]:
    margin_left = 10 * (item_list[-1][1] - 1)
    style = Style(f"margin-left: {margin_left}px")
    if type == "bulleted_list":
        html.append(Ul([style], [item[2] for item in item_list]))
    else:
        html.append(Ol([style], [item[2] for item in item_list]))
    return [], html


def build_html(
    current_block: dict, bulleted_list: list, numbered_list: list
) -> Tuple[list, list, list]:
    html = []
    # extract current block's html
    if isinstance(current_block["block"].block, notion_blocks.BulletedListItem):
        if bulleted_list and current_block["parent_id"] != bulleted_list[-1][0]:
            bulleted_list, html = flush_list("bulleted_list", bulleted_list, html)
        bulleted_list.append(
            (current_block["parent_id"], current_block["level"], current_block["block"].get_html())
        )
        if bulleted_list and current_block["peers_rank"] == current_block["peers_count"] - 1:
            bulleted_list, html = flush_list("bulleted_list", bulleted_list, html)
    elif isinstance(current_block["block"].block, notion_blocks.NumberedListItem):
        if numbered_list and current_block["parent_id"] != numbered_list[-1][0]:
            numbered_list, html = flush_list("numbered_list", numbered_list, html)
        numbered_list.append(
            (current_block["parent_id"], current_block["level"], current_block["block"].get_html())
        )
        if numbered_list and current_block["peers_rank"] == current_block["peers_count"] - 1:
            numbered_list, html = flush_list("numbered_list", numbered_list, html)
    else:
        if bulleted_list:
            bulleted_list, html = flush_list("bulleted_list", bulleted_list, html)
        if numbered_list:
            numbered_list, html = flush_list("numbered_list", numbered_list, html)
        if (
            isinstance(current_block["block"].block, notion_blocks.TableRow)
            and current_block["peers_rank"] == 0
        ):
            current_block["block"].is_header = True
        if current_block["block"].get_html():
            html.append(current_block["block"].get_html())
        else:
            html.append([])
    # process current block's children
    if current_block["children"]:
        children_html = []
        for index, child in enumerate(current_block["children"]):
            if child:
                child["peers_rank"] = index
                child["peers_count"] = len(current_block["children"])
                child_html, bulleted_list, numbered_list = build_html(
                    child, bulleted_list, numbered_list
                )
                if child_html:
                    children_html.append(child_html)
        if isinstance(current_block["block"].block, notion_blocks.Column):
            html.append(
                Div(
                    [Style(f"width:{100 / current_block['peers_count']}%; float: left")],
                    children_html,
                )
            )
        elif isinstance(current_block["block"].block, notion_blocks.Table):
            html.append(Table([], children_html))
        else:
            html.append(Div([], children_html))

    return html, bulleted_list, numbered_list


def extract_page_html(
    client: Client,
    page_id: str,
    logger: logging.Logger,
) -> HtmlExtractionResponse:
    parent_page_id = UUID(page_id)
    parent_block: Block = client.blocks.retrieve(block_id=page_id)  # type: ignore
    head = None
    if isinstance(parent_block.block, notion_blocks.ChildPage):
        head = Head([], Title([], parent_block.block.title))
    current_block = {
        "block": parent_block,
        "level": 0,
        "children": [],
        "parent_id": None,
        "peers_rank": 0,
        "peers_count": 1,
    }
    logger.debug(f"processing page id: {page_id}")
    current_block, child_pages, child_databases = process_block(
        current_block, parent_page_id, client, [], []
    )
    html, _, _ = build_html(current_block, [], [])
    body = Body([], html)
    all_elements = [body]
    if head:
        all_elements = [head] + all_elements
    full_html = Html([], all_elements)
    return HtmlExtractionResponse(
        full_html,
        child_pages=child_pages,
        child_databases=child_databases,
    )


def extract_database_html(
    client: Client,
    database_id: str,
    logger: logging.Logger,
) -> HtmlExtractionResponse:
    logger.debug(f"processing database id: {database_id}")
    database: Database = client.databases.retrieve(database_id=database_id)  # type: ignore
    property_keys = list(database.properties.keys())
    property_keys = sorted(property_keys)
    table_html_rows = []
    child_pages: List[str] = []
    child_databases: List[str] = []
    # Create header row
    table_html_rows.append(Tr([], [Th([], k) for k in property_keys]))

    all_pages = []
    for page_chunk in client.databases.iterate_query(database_id=database_id):  # type: ignore
        all_pages.extend(page_chunk)

    logger.debug(f"creating {len(all_pages)} rows")
    for page in all_pages:
        if is_database_url(client=client, url=page.url):
            child_databases.append(page.id)
        if is_page_url(client=client, url=page.url):
            child_pages.append(page.id)
        properties = page.properties
        inner_html = [properties.get(k).get_html() for k in property_keys]  # type: ignore
        table_html_rows.append(
            Tr(
                [],
                [Td([], cell) for cell in [html if html else Div([], []) for html in inner_html]],
            ),
        )

    table_html = Table([], table_html_rows)

    return HtmlExtractionResponse(
        html=table_html,
        child_pages=child_pages,
        child_databases=child_databases,
    )


@dataclass
class ChildExtractionResponse:
    child_pages: List[str] = field(default_factory=list)
    child_databases: List[str] = field(default_factory=list)


class QueueEntryType(enum.Enum):
    DATABASE = "database"
    PAGE = "page"


@dataclass
class QueueEntry:
    type: QueueEntryType
    id: UUID


def get_recursive_content_from_page(
    client: Client,
    page_id: str,
    logger: logging.Logger,
) -> ChildExtractionResponse:
    return get_recursive_content(
        client=client,
        init_entry=QueueEntry(type=QueueEntryType.PAGE, id=UUID(page_id)),
        logger=logger,
    )


def get_recursive_content_from_database(
    client: Client,
    database_id: str,
    logger: logging.Logger,
) -> ChildExtractionResponse:
    return get_recursive_content(
        client=client,
        init_entry=QueueEntry(type=QueueEntryType.DATABASE, id=UUID(database_id)),
        logger=logger,
    )


def get_recursive_content(
    client: Client,
    init_entry: QueueEntry,
    logger: logging.Logger,
) -> ChildExtractionResponse:
    parents: List[QueueEntry] = [init_entry]
    child_pages: List[str] = []
    child_dbs: List[str] = []
    processed: List[str] = []
    while len(parents) > 0:
        parent: QueueEntry = parents.pop()
        processed.append(str(parent.id))
        if parent.type == QueueEntryType.PAGE:
            logger.debug(f"getting child data from page: {parent.id}")
            page_children = []
            try:
                for children_block in client.blocks.children.iterate_list(  # type: ignore
                    block_id=str(parent.id),
                ):
                    page_children.extend(children_block)
            except APIResponseError as api_error:
                logger.error(f"failed to get page with id {parent.id}: {api_error}")
                if str(parent.id) in child_pages:
                    child_pages.remove(str(parent.id))
                continue
            if not page_children:
                continue

            # Extract child pages
            child_pages_from_page = [
                c for c in page_children if isinstance(c.block, notion_blocks.ChildPage)
            ]
            if child_pages_from_page:
                child_page_blocks: List[notion_blocks.ChildPage] = [
                    p.block
                    for p in child_pages_from_page
                    if isinstance(p.block, notion_blocks.ChildPage)
                ]
                logger.debug(
                    "found child pages from parent page {}: {}".format(
                        parent.id,
                        ", ".join([block.title for block in child_page_blocks]),
                    ),
                )
            new_pages = [p.id for p in child_pages_from_page if p.id not in processed]
            new_pages = list(set(new_pages))
            child_pages.extend(new_pages)
            parents.extend(
                [QueueEntry(type=QueueEntryType.PAGE, id=UUID(i)) for i in new_pages],
            )

            # Extract child databases
            child_dbs_from_page = [
                c for c in page_children if isinstance(c.block, notion_blocks.ChildDatabase)
            ]
            if child_dbs_from_page:
                child_db_blocks: List[notion_blocks.ChildDatabase] = [
                    c.block
                    for c in page_children
                    if isinstance(c.block, notion_blocks.ChildDatabase)
                ]
                logger.debug(
                    "found child database from parent page {}: {}".format(
                        parent.id,
                        ", ".join([block.title for block in child_db_blocks]),
                    ),
                )
            new_dbs = [db.id for db in child_dbs_from_page if db.id not in processed]
            new_dbs = list(set(new_dbs))
            child_dbs.extend(new_dbs)
            parents.extend(
                [QueueEntry(type=QueueEntryType.DATABASE, id=UUID(i)) for i in new_dbs],
            )

            linked_to_others: List[notion_blocks.LinkToPage] = [
                c.block for c in page_children if isinstance(c.block, notion_blocks.LinkToPage)
            ]
            for link in linked_to_others:
                if (page_id := link.page_id) and (
                    page_id not in processed and page_id not in child_pages
                ):
                    child_pages.append(page_id)
                    parents.append(QueueEntry(type=QueueEntryType.PAGE, id=UUID(page_id)))
                if (database_id := link.database_id) and (
                    database_id not in processed and database_id not in child_dbs
                ):
                    child_dbs.append(database_id)
                    parents.append(
                        QueueEntry(type=QueueEntryType.DATABASE, id=UUID(database_id)),
                    )

        elif parent.type == QueueEntryType.DATABASE:
            logger.debug(f"getting child data from database: {parent.id}")
            database_pages = []
            try:
                for page_entries in client.databases.iterate_query(  # type: ignore
                    database_id=str(parent.id),
                ):
                    database_pages.extend(page_entries)
            except APIResponseError as api_error:
                logger.error(f"failed to get database with id {parent.id}: {api_error}")
                if str(parent.id) in child_dbs:
                    child_dbs.remove(str(parent.id))
                continue
            if not database_pages:
                continue

            child_pages_from_db = [
                p for p in database_pages if is_page_url(client=client, url=p.url)
            ]
            if child_pages_from_db:
                logger.debug(
                    "found child pages from parent database {}: {}".format(
                        parent.id,
                        ", ".join([p.url for p in child_pages_from_db]),
                    ),
                )
            new_pages = [p.id for p in child_pages_from_db if p.id not in processed]
            child_pages.extend(new_pages)
            parents.extend(
                [QueueEntry(type=QueueEntryType.PAGE, id=UUID(i)) for i in new_pages],
            )

            child_dbs_from_db = [
                p for p in database_pages if is_database_url(client=client, url=p.url)
            ]
            if child_dbs_from_db:
                logger.debug(
                    "found child database from parent database {}: {}".format(
                        parent.id,
                        ", ".join([db.url for db in child_dbs_from_db]),
                    ),
                )
            new_dbs = [db.id for db in child_dbs_from_db if db.id not in processed]
            child_dbs.extend(new_dbs)
            parents.extend(
                [QueueEntry(type=QueueEntryType.DATABASE, id=UUID(i)) for i in new_dbs],
            )

    return ChildExtractionResponse(
        child_pages=child_pages,
        child_databases=child_dbs,
    )


def is_valid_uuid(uuid_str: str) -> bool:
    try:
        UUID(uuid_str)
        return True
    except Exception:
        return False


def get_uuid_from_url(path: str) -> Optional[str]:
    strings = path.split("-")
    if len(strings) > 0 and is_valid_uuid(strings[-1]):
        return strings[-1]
    return None


def is_page_url(client: Client, url: str):
    parsed_url = urlparse(url)
    path = parsed_url.path.split("/")[-1]
    if parsed_url.netloc != "www.notion.so":
        return False
    page_uuid = get_uuid_from_url(path=path)
    if not page_uuid:
        return False
    check_resp = client.pages.retrieve_status(page_id=page_uuid)
    return check_resp == 200


def is_database_url(client: Client, url: str):
    parsed_url = urlparse(url)
    path = parsed_url.path.split("/")[-1]
    if parsed_url.netloc != "www.notion.so":
        return False
    database_uuid = get_uuid_from_url(path=path)
    if not database_uuid:
        return False
    check_resp = client.databases.retrieve_status(database_id=database_uuid)
    return check_resp == 200



from __future__ import annotations

import asyncio
import uuid
from collections import defaultdict
from contextlib import asynccontextmanager
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, AsyncGenerator, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field, Secret, ValidationError, field_validator

from unstructured_ingest.data_types.entities import EntitiesData, Entity, EntityRelationship
from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.error import (
    DestinationConnectionError,
    UnstructuredIngestError,
    ValueError,
)
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
)
from unstructured_ingest.processes.connectors.utils import format_and_truncate_orig_elements
from unstructured_ingest.utils.data_prep import batch_generator, get_json_data
from unstructured_ingest.utils.dep_check import requires_dependencies

SimilarityFunction = Literal["cosine"]

if TYPE_CHECKING:
    from neo4j import AsyncDriver, Auth
    from networkx import Graph, MultiDiGraph

CONNECTOR_TYPE = "neo4j"


class Neo4jAccessConfig(AccessConfig):
    password: str


class Neo4jConnectionConfig(ConnectionConfig):
    access_config: Secret[Neo4jAccessConfig]
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)
    username: str = Field(default="neo4j")
    uri: str = Field(description="Neo4j Connection URI <scheme>://<host>:<port>")
    database: str = Field(default="neo4j", description="Name of the target database")

    @requires_dependencies(["neo4j"], extras="neo4j")
    @asynccontextmanager
    async def get_client(self) -> AsyncGenerator["AsyncDriver", None]:
        from neo4j import AsyncGraphDatabase

        driver = AsyncGraphDatabase.driver(**self._get_driver_parameters())
        logger.info(f"Created driver connecting to the database '{self.database}' at {self.uri}.")
        try:
            yield driver
        finally:
            await driver.close()
            logger.info(
                f"Closed driver connecting to the database '{self.database}' at {self.uri}."
            )

    def _get_driver_parameters(self) -> dict:
        return {
            "uri": self.uri,
            "auth": self._get_auth(),
            "database": self.database,
        }

    @requires_dependencies(["neo4j"], extras="neo4j")
    def _get_auth(self) -> "Auth":
        from neo4j import Auth

        return Auth("basic", self.username, self.access_config.get_secret_value().password)


class Neo4jUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class Neo4jUploadStager(UploadStager):
    upload_stager_config: Neo4jUploadStagerConfig = Field(
        default_factory=Neo4jUploadStagerConfig, validate_default=True
    )

    def run(  # type: ignore
        self,
        elements_filepath: Path,
        file_data: FileData,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        elements = get_json_data(elements_filepath)
        nx_graph = self._create_lexical_graph(
            elements, self._create_document_node(file_data=file_data)
        )
        output_filepath = Path(output_dir) / f"{output_filename}.json"
        output_filepath.parent.mkdir(parents=True, exist_ok=True)

        with open(output_filepath, "w") as file:
            file.write(_GraphData.from_nx(nx_graph).model_dump_json())

        return output_filepath

    def _add_entities(self, entities: list[Entity], graph: "Graph", element_node: _Node) -> None:
        for entity in entities:
            entity_node = _Node(
                labels=[Label.ENTITY], properties={"id": entity.entity}, id_=entity.entity
            )
            graph.add_edge(
                entity_node,
                _Node(labels=[Label.ENTITY], properties={"id": entity.type}, id_=entity.type),
                relationship=Relationship.ENTITY_TYPE,
            )
            graph.add_edge(element_node, entity_node, relationship=Relationship.HAS_ENTITY)

    def _add_entity_relationships(
        self, relationships: list[EntityRelationship], graph: "Graph"
    ) -> None:
        for relationship in relationships:
            from_node = _Node(
                labels=[Label.ENTITY],
                properties={"id": relationship.from_},
                id_=relationship.from_,
            )
            to_node = _Node(
                labels=[Label.ENTITY], properties={"id": relationship.to}, id_=relationship.to
            )
            graph.add_edge(from_node, to_node, relationship=relationship.relationship)

    def _add_entity_data(self, element: dict, graph: "Graph", element_node: _Node) -> None:
        entities = element.get("metadata", {}).get("entities", {})
        if not entities:
            return None
        try:
            if isinstance(entities, list):
                self._add_entities(
                    [Entity.model_validate(e) for e in entities if isinstance(e, dict)],
                    graph,
                    element_node,
                )
            elif isinstance(entities, dict):
                entity_data = EntitiesData.model_validate(entities)
                self._add_entities(entity_data.items, graph, element_node)
                self._add_entity_relationships(entity_data.relationships, graph)
        except ValidationError:
            logger.warning(
                "Failed to add entities to the graph. "
                "Please check the format of the entities in the input data."
            )
        return None

    def _create_lexical_graph(self, elements: list[dict], document_node: _Node) -> "Graph":
        import networkx as nx

        graph = nx.MultiDiGraph()
        graph.add_node(document_node)

        previous_node: Optional[_Node] = None
        for element in elements:
            element_node = self._create_element_node(element)
            order_relationship = (
                Relationship.NEXT_CHUNK if self._is_chunk(element) else Relationship.NEXT_ELEMENT
            )
            if previous_node:
                graph.add_edge(element_node, previous_node, relationship=order_relationship)

            previous_node = element_node
            graph.add_edge(element_node, document_node, relationship=Relationship.PART_OF_DOCUMENT)

            self._add_entity_data(element, graph, element_node)

            if self._is_chunk(element):
                for origin_element in format_and_truncate_orig_elements(element, include_text=True):
                    origin_element_node = self._create_element_node(origin_element)

                    graph.add_edge(
                        origin_element_node,
                        element_node,
                        relationship=Relationship.PART_OF_CHUNK,
                    )
                    graph.add_edge(
                        origin_element_node,
                        document_node,
                        relationship=Relationship.PART_OF_DOCUMENT,
                    )
                    self._add_entity_data(origin_element, graph, origin_element_node)

        return graph

    # TODO(Filip Knefel): Ensure _is_chunk is as reliable as possible, consider different checks
    def _is_chunk(self, element: dict) -> bool:
        return "orig_elements" in element.get("metadata", {})

    def _create_document_node(self, file_data: FileData) -> _Node:
        properties = {}
        if file_data.source_identifiers:
            properties["name"] = file_data.source_identifiers.filename
        if file_data.metadata.date_created:
            properties["date_created"] = file_data.metadata.date_created
        if file_data.metadata.date_modified:
            properties["date_modified"] = file_data.metadata.date_modified
        return _Node(id_=file_data.identifier, properties=properties, labels=[Label.DOCUMENT])

    def _create_element_node(self, element: dict) -> _Node:
        properties = {"id": element["element_id"]}

        if text := element.get("text"):
            # if we have chunks, we won't have text here for the original elements
            properties["text"] = text

        if embeddings := element.get("embeddings"):
            properties["embeddings"] = embeddings

        label = Label.CHUNK if self._is_chunk(element) else Label.UNSTRUCTURED_ELEMENT
        return _Node(id_=element["element_id"], properties=properties, labels=[label])


class _GraphData(BaseModel):
    nodes: list[_Node]
    edges: list[_Edge]

    @classmethod
    def from_nx(cls, nx_graph: "MultiDiGraph") -> _GraphData:
        nodes = list(nx_graph.nodes())
        edges = [
            _Edge(
                source=u,
                destination=v,
                relationship=Relationship(data_dict["relationship"])
                if data_dict["relationship"] in Relationship
                else data_dict["relationship"],
            )
            for u, v, data_dict in nx_graph.edges(data=True)
        ]
        return _GraphData(nodes=nodes, edges=edges)


class _Node(BaseModel):
    model_config = ConfigDict()

    labels: list[Label]
    properties: dict = Field(default_factory=dict)
    id_: str = Field(default_factory=lambda: str(uuid.uuid4()))

    def __hash__(self):
        return hash(self.id_)

    @property
    def main_label(self) -> Label:
        return self.labels[0]

    @classmethod
    @field_validator("labels", mode="after")
    def require_at_least_one_label(cls, value: list[Label]) -> list[Label]:
        if not value:
            raise ValueError("Node must have at least one label.")
        return value


class _Edge(BaseModel):
    model_config = ConfigDict()

    source: _Node
    destination: _Node
    relationship: Relationship | str


class Label(Enum):
    UNSTRUCTURED_ELEMENT = "UnstructuredElement"
    CHUNK = "Chunk"
    DOCUMENT = "Document"
    ENTITY = "Entity"


class Relationship(Enum):
    PART_OF_DOCUMENT = "PART_OF_DOCUMENT"
    PART_OF_CHUNK = "PART_OF_CHUNK"
    NEXT_CHUNK = "NEXT_CHUNK"
    NEXT_ELEMENT = "NEXT_ELEMENT"
    ENTITY_TYPE = "ENTITY_TYPE"
    HAS_ENTITY = "HAS_ENTITY"


class Neo4jUploaderConfig(UploaderConfig):
    batch_size: int = Field(
        default=1000, description="Maximal number of nodes/relationships created per transaction."
    )
    similarity_function: SimilarityFunction = Field(
        default="cosine",
        description="Vector similarity function used to create index on Chunk nodes",
    )
    create_destination: bool = Field(
        default=True, description="Create destination if it does not exist"
    )


@dataclass
class Neo4jUploader(Uploader):
    upload_config: Neo4jUploaderConfig
    connection_config: Neo4jConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    @DestinationConnectionError.wrap
    def precheck(self) -> None:
        async def verify_auth():
            async with self.connection_config.get_client() as client:
                await client.verify_connectivity()

        asyncio.run(verify_auth())

    def is_async(self):
        return True

    async def run_async(self, path: Path, file_data: FileData, **kwargs) -> None:  # type: ignore
        staged_data = get_json_data(path)

        graph_data = _GraphData.model_validate(staged_data)
        async with self.connection_config.get_client() as client:
            await self._create_uniqueness_constraints(client)
            embedding_dimensions = self._get_embedding_dimensions(graph_data)
            if embedding_dimensions and self.upload_config.create_destination:
                await self._create_vector_index(
                    client,
                    dimensions=embedding_dimensions,
                    similarity_function=self.upload_config.similarity_function,
                )
            await self._delete_old_data_if_exists(file_data, client=client)
            await self._merge_graph(graph_data=graph_data, client=client)

    async def _create_uniqueness_constraints(self, client: AsyncDriver) -> None:
        for label in Label:
            logger.info(
                f"Adding id uniqueness constraint for nodes labeled '{label.value}'"
                " if it does not already exist."
            )
            constraint_name = f"{label.value.lower()}_id"
            await client.execute_query(
                f"""
                CREATE CONSTRAINT {constraint_name} IF NOT EXISTS
                FOR (n: {label.value}) REQUIRE n.id IS UNIQUE
                """
            )

    async def _create_vector_index(
        self, client: AsyncDriver, dimensions: int, similarity_function: SimilarityFunction
    ) -> None:
        import neo4j.exceptions

        label = Label.CHUNK
        logger.info(
            f"Creating index on nodes labeled '{label.value}' if it does not already exist."
        )
        index_name = f"{label.value.lower()}_vector"
        try:
            await client.execute_query(
                f"""
            CREATE VECTOR INDEX {index_name} IF NOT EXISTS
            FOR (n:{label.value}) ON n.embedding
            OPTIONS {{indexConfig: {{
                `vector.similarity_function`: '{similarity_function}',
                `vector.dimensions`: {dimensions}}}
            }}
            """
            )
        except neo4j.exceptions.ClientError as e:
            if e.code == "Neo.ClientError.Schema.EquivalentSchemaRuleAlreadyExists":
                logger.info(f"Index on nodes labeled '{label.value}' already exists.")
            else:
                raise UnstructuredIngestError(str(e))

    async def _delete_old_data_if_exists(self, file_data: FileData, client: AsyncDriver) -> None:
        logger.info(f"Deleting old data for the record '{file_data.identifier}' (if present).")
        _, summary, _ = await client.execute_query(
            f"""
            MATCH (n: `{Label.DOCUMENT.value}` {{id: $identifier}})
            MATCH (n)--(m: `{Label.CHUNK.value}`|`{Label.UNSTRUCTURED_ELEMENT.value}`)
            DETACH DELETE m
            DETACH DELETE n""",
            identifier=file_data.identifier,
        )
        logger.info(
            f"Deleted {summary.counters.nodes_deleted} nodes"
            f" and {summary.counters.relationships_deleted} relationships."
        )

    async def _merge_graph(self, graph_data: _GraphData, client: AsyncDriver) -> None:
        nodes_by_labels: defaultdict[Label, list[_Node]] = defaultdict(list)
        for node in graph_data.nodes:
            nodes_by_labels[node.main_label].append(node)
        logger.info(f"Merging {len(graph_data.nodes)} graph nodes.")
        # NOTE: Processed in parallel as there's no overlap between accessed nodes
        await self._execute_queries(
            [
                self._create_nodes_query(nodes_batch, label)
                for label, nodes in nodes_by_labels.items()
                for nodes_batch in batch_generator(nodes, batch_size=self.upload_config.batch_size)
            ],
            client=client,
            in_parallel=True,
        )
        logger.info(f"Finished merging {len(graph_data.nodes)} graph nodes.")

        edges_by_relationship: defaultdict[tuple[Relationship | str, Label, Label], list[_Edge]] = (
            defaultdict(list)
        )
        for edge in graph_data.edges:
            key = (edge.relationship, edge.source.main_label, edge.destination.main_label)
            edges_by_relationship[key].append(edge)

        logger.info(f"Merging {len(graph_data.edges)} graph relationships (edges).")
        # NOTE: Processed sequentially to avoid queries locking node access to one another
        await self._execute_queries(
            [
                self._create_edges_query(edges_batch, relationship, source_label, destination_label)
                for (
                    relationship,
                    source_label,
                    destination_label,
                ), edges in edges_by_relationship.items()
                for edges_batch in batch_generator(edges, batch_size=self.upload_config.batch_size)
            ],
            client=client,
        )
        logger.info(f"Finished merging {len(graph_data.edges)} graph relationships (edges).")

    @staticmethod
    async def _execute_queries(
        queries_with_parameters: list[tuple[str, dict]],
        client: AsyncDriver,
        in_parallel: bool = False,
    ) -> None:
        from neo4j import EagerResult

        results: list[EagerResult] = []
        logger.info(
            f"Executing {len(queries_with_parameters)} "
            + f"{'parallel' if in_parallel else 'sequential'} Cypher statements."
        )
        if in_parallel:
            results = await asyncio.gather(
                *[
                    client.execute_query(query, parameters_=parameters)
                    for query, parameters in queries_with_parameters
                ]
            )
        else:
            for i, (query, parameters) in enumerate(queries_with_parameters):
                logger.info(f"Statement #{i} started.")
                results.append(await client.execute_query(query, parameters_=parameters))
                logger.info(f"Statement #{i} finished.")
        nodeCount = sum([res.summary.counters.nodes_created for res in results])
        relCount = sum([res.summary.counters.relationships_created for res in results])
        logger.info(
            f"Finished executing all ({len(queries_with_parameters)}) "
            + f"{'parallel' if in_parallel else 'sequential'} Cypher statements. "
            + f"Created {nodeCount} nodes, {relCount} relationships."
        )

    @staticmethod
    def _create_nodes_query(nodes: list[_Node], label: Label) -> tuple[str, dict]:
        logger.info(f"Preparing MERGE query for {len(nodes)} nodes labeled '{label}'.")
        query_string = f"""
            UNWIND $nodes AS node
            MERGE (n: `{label.value}` {{id: node.id}})
            SET n += node.properties
            SET n:$(node.labels)
            WITH * WHERE node.vector IS NOT NULL
            CALL db.create.setNodeVectorProperty(n, 'embedding', node.vector)
            """
        parameters = {
            "nodes": [
                {
                    "id": node.id_,
                    "labels": [l.value for l in node.labels if l != label],  # noqa: E741
                    "vector": node.properties.pop("embedding", None),
                    "properties": node.properties,
                }
                for node in nodes
            ]
        }
        return query_string, parameters

    @staticmethod
    def _create_edges_query(
        edges: list[_Edge],
        relationship: Relationship | str,
        source_label: Label,
        destination_label: Label,
    ) -> tuple[str, dict]:
        logger.info(f"Preparing MERGE query for {len(edges)} {relationship} relationships.")
        relationship = (
            relationship.value if isinstance(relationship, Relationship) else relationship
        )
        query_string = f"""
            UNWIND $edges AS edge
            MATCH (u: `{source_label.value}` {{id: edge.source}})
            MATCH (v: `{destination_label.value}` {{id: edge.destination}})
            MERGE (u)-[:`{relationship}`]->(v)
            """
        parameters = {
            "edges": [
                {"source": edge.source.id_, "destination": edge.destination.id_} for edge in edges
            ]
        }
        return query_string, parameters

    def _get_embedding_dimensions(self, graph_data: _GraphData) -> int | None:
        """Embedding dimensions inferred from chunk nodes or None if it can't be determined."""
        for node in graph_data.nodes:
            if Label.CHUNK in node.labels and "embeddings" in node.properties:
                return len(node.properties["embeddings"])

        return None


neo4j_destination_entry = DestinationRegistryEntry(
    connection_config=Neo4jConnectionConfig,
    upload_stager=Neo4jUploadStager,
    upload_stager_config=Neo4jUploadStagerConfig,
    uploader=Neo4jUploader,
    uploader_config=Neo4jUploaderConfig,
)



from __future__ import annotations

from unstructured_ingest.processes.connector_registry import (
    add_destination_entry,
    add_source_entry,
)

from .volumes_aws import CONNECTOR_TYPE as VOLUMES_AWS_CONNECTOR_TYPE
from .volumes_aws import (
    databricks_aws_volumes_destination_entry,
    databricks_aws_volumes_source_entry,
)
from .volumes_azure import CONNECTOR_TYPE as VOLUMES_AZURE_CONNECTOR_TYPE
from .volumes_azure import (
    databricks_azure_volumes_destination_entry,
    databricks_azure_volumes_source_entry,
)
from .volumes_gcp import CONNECTOR_TYPE as VOLUMES_GCP_CONNECTOR_TYPE
from .volumes_gcp import (
    databricks_gcp_volumes_destination_entry,
    databricks_gcp_volumes_source_entry,
)
from .volumes_native import CONNECTOR_TYPE as VOLUMES_NATIVE_CONNECTOR_TYPE
from .volumes_native import (
    databricks_native_volumes_destination_entry,
    databricks_native_volumes_source_entry,
)
from .volumes_table import CONNECTOR_TYPE as VOLUMES_TABLE_CONNECTOR_TYPE
from .volumes_table import databricks_volumes_delta_tables_destination_entry

add_source_entry(source_type=VOLUMES_AWS_CONNECTOR_TYPE, entry=databricks_aws_volumes_source_entry)
add_destination_entry(
    destination_type=VOLUMES_AWS_CONNECTOR_TYPE, entry=databricks_aws_volumes_destination_entry
)

add_source_entry(source_type=VOLUMES_GCP_CONNECTOR_TYPE, entry=databricks_gcp_volumes_source_entry)
add_destination_entry(
    destination_type=VOLUMES_GCP_CONNECTOR_TYPE, entry=databricks_gcp_volumes_destination_entry
)

add_source_entry(
    source_type=VOLUMES_NATIVE_CONNECTOR_TYPE, entry=databricks_native_volumes_source_entry
)
add_destination_entry(
    destination_type=VOLUMES_NATIVE_CONNECTOR_TYPE,
    entry=databricks_native_volumes_destination_entry,
)

add_source_entry(
    source_type=VOLUMES_AZURE_CONNECTOR_TYPE, entry=databricks_azure_volumes_source_entry
)
add_destination_entry(
    destination_type=VOLUMES_AZURE_CONNECTOR_TYPE, entry=databricks_azure_volumes_destination_entry
)
add_destination_entry(
    destination_type=VOLUMES_TABLE_CONNECTOR_TYPE,
    entry=databricks_volumes_delta_tables_destination_entry,
)
